from ldp.nn.handlers.chunking import TensorChunker

from .agent.simple_local_agent import AgentLMConfig, SimpleLocalLLMAgent
from .graph.llm_call_op import LocalLLMCallOp
from .handlers.transformer_handler import (
    AsyncTransformer,
    AsyncTransformerInterface,
    ExecutionMode,
    LMType,
    ParallelAsyncTransformer,
    ParallelModeConfig,
    ParallelTransformerHandler,
    TransformerHandler,
    TransformerHandlerConfig,
    collate_fn_transformer_left_pad,
    collate_fn_transformer_right_pad,
    decollate_fn_transformer_decoder,
)
from .lm_config import LMConfig, TorchDType
from .utils import set_seed

__all__ = [
    "AgentLMConfig",
    "AsyncTransformer",
    "AsyncTransformerInterface",
    "ExecutionMode",
    "LMConfig",
    "LMType",
    "LocalLLMCallOp",
    "ParallelAsyncTransformer",
    "ParallelModeConfig",
    "ParallelTransformerHandler",
    "SimpleLocalLLMAgent",
    "TensorChunker",
    "TorchDType",
    "TransformerHandler",
    "TransformerHandlerConfig",
    "collate_fn_transformer_left_pad",
    "collate_fn_transformer_right_pad",
    "decollate_fn_transformer_decoder",
    "set_seed",
]



from __future__ import annotations

import json
from functools import partial
from typing import TYPE_CHECKING, Any, ClassVar

from aviary.core import MalformedMessageError, Message, Messages
from aviary.tools import Tool, ToolCall, ToolRequestMessage
from transformers import LogitsProcessorList

from ldp.graph.gradient_estimators import assign_constant_grads
from ldp.graph.op_utils import CallID, get_call_id, get_training_mode
from ldp.graph.ops import GradInType, Op, OpCtx, ResultOrValue
from ldp.nn.handlers.transformer_handler import (
    AsyncTransformerInterface,
    LMType,
    ParallelModeConfig,
    TransformerHandlerConfig,
    collate_fn_transformer_left_pad,
    decollate_fn_transformer_decoder,
)
from ldp.nn.lm_config import LMConfig

if TYPE_CHECKING:
    import tree


class LocalLLMCallOp(Op[Message]):
    """An Op that samples a token sequence from a local language model."""

    CTX_INPUTS_PREP_KEY: ClassVar[str] = "inputs_prepared"
    CTX_TOOLS_PREP_KEY: ClassVar[str] = "tools_prepared"
    CTX_OUTPUT_PREP_KEY: ClassVar[str] = "outputs_prepared"

    model_name: str

    def __init__(
        self,
        model_config: LMConfig,
        batch_size: int = 1,
        max_wait_interval: float = 0.1,
        parallel_mode_config: ParallelModeConfig | None = None,
    ) -> None:
        super().__init__()

        pad_token_id = model_config.get_tokenizer().pad_token_id

        handler_config = TransformerHandlerConfig(
            # configurable
            lm_config=model_config,
            batch_size=batch_size,
            max_wait_interval=max_wait_interval,
            parallel_mode_config=parallel_mode_config,
            # constant configuration
            lm_type=LMType.GENERATION,
            module_call_fn=AsyncTransformerInterface.model_generate,
            collate_fn=partial(
                collate_fn_transformer_left_pad, pad_token_id=pad_token_id
            ),
            decollate_fn=decollate_fn_transformer_decoder,
        )
        self.model_handler = handler_config.make_async_module()
        self.model_name = model_config.model

        self.llm_call_kwargs = {"logits_processor": LogitsProcessorList()}

    @staticmethod
    def prep_messages_for_tokenizer(xi: Messages) -> list[dict]:
        result: list[dict] = []
        for msg in xi:
            content = msg.content
            if isinstance(msg, ToolRequestMessage):
                assert len(msg.tool_calls) == 1, (
                    "Support parsing only single tool call for now"
                )
                tool_call = msg.tool_calls[0]
                # TODO: document where this format is coming from. Is this a Huggingface chat template syntax?
                content_dict = {
                    "name": tool_call.function.name,
                    "parameters": tool_call.function.arguments,
                    "thought": msg.content,
                }
                content = json.dumps(content_dict)
            assert content is not None, "content is None, doesn't make sense"

            result.append({"role": msg.role, "content": content})
        return result

    @staticmethod
    def prep_tools_for_tokenizer(
        tools: list[Tool] | None,
    ) -> list[dict[Any, Any]] | None:
        """Prepare tools for the tokenizer by transforming them into a JSON schema."""
        if not tools:
            return None

        # TODO: should be able to switch to tool.info.model_dump() here
        tools_list: list[dict[Any, Any]] = []
        for tool in tools:
            if tool.info.parameters is None:
                raise NotImplementedError(
                    "Didn't yet handle serializing tools without parameters."
                )

            tools_list.append({
                "name": tool.info.name,
                "description": tool.info.description,
                "parameters": {
                    "type": tool.info.parameters.type,
                    "properties": {
                        prop_name: {
                            "type": prop_details.get("type"),
                            "description": prop_details.get("description"),
                            "title": prop_details.get("title"),
                        }
                        for prop_name, prop_details in tool.info.get_properties().items()
                    },
                    "required": tool.info.parameters.required,
                },
            })
        return tools_list

    @staticmethod
    def _parse_tool_request(out_text: str, tools: list[Tool]) -> ToolRequestMessage:
        """Parse the output text to extract the tool request.

        TODO: see if this needs to be configurable, e.g. for different model
        output formats that we want to experiment with.
        """
        try:
            tool_request = json.loads(out_text)
            tool_name = tool_request["name"]
            tool = next(t for t in tools if t.info.name == tool_name)
            tool_thought = tool_request.get("thought", "")
            tool_parameters = tool_request.get("parameters", {})
            return ToolRequestMessage(
                tool_calls=[ToolCall.from_tool(tool, **tool_parameters)],
                content=tool_thought,
            )
        except StopIteration as exc:
            raise MalformedMessageError(
                f"Tool {tool_name} not found in tools."
            ) from exc
        except json.JSONDecodeError as err:
            raise ValueError(f"Failed to parse tools call message: {out_text}") from err

    async def forward(
        self,
        xi: list[Message],
        temperature: float = 1.0,
        max_new_tokens: int = 10,
        tools: list[Tool] | None = None,
        **kwargs: dict[str, Any],
    ) -> Message:
        call_id = get_call_id()
        inputs = self.prep_messages_for_tokenizer(xi)
        tools_json = self.prep_tools_for_tokenizer(tools)
        if get_training_mode():
            self.ctx.update(call_id, LocalLLMCallOp.CTX_INPUTS_PREP_KEY, inputs)
            self.ctx.update(call_id, LocalLLMCallOp.CTX_TOOLS_PREP_KEY, tools_json)

        out_text, logprobs = await self.model_handler(
            inputs,
            temperature=temperature,
            max_new_tokens=max_new_tokens,
            tools_json=tools_json,
            output_scores=True,
            return_legacy_cache=False,
            return_dict_in_generate=True,
            do_sample=temperature > 0,
            **self.llm_call_kwargs,
            **kwargs,
        )

        out_msg = Message(role="assistant", content=out_text)
        if tools and out_text.startswith("{"):
            out_msg = self._parse_tool_request(out_text, tools)

        if get_training_mode():
            self.ctx.update(
                call_id,
                LocalLLMCallOp.CTX_OUTPUT_PREP_KEY,
                self.prep_messages_for_tokenizer([out_msg])[0],
            )
            self.ctx.update(call_id, "logprob", logprobs.cpu().tolist())

        return out_msg

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        return assign_constant_grads(input_args, input_kwargs, 0.0, descend=False)



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable

from ldp.data_structures import Trajectory
from ldp.shims import tqdm

logger = logging.getLogger(__name__)


# Registry for all optimizers
_OPTIMIZER_REGISTRY: dict[str, type[Optimizer]] = {}


class Optimizer(ABC):
    """Base class for all optimizers."""

    def __init_subclass__(cls) -> None:
        # Register each optimizer subclass
        _OPTIMIZER_REGISTRY[cls.__name__] = cls
        return super().__init_subclass__()

    def aggregate(
        self, trajectories: Iterable[Trajectory], show_pbar: bool = False, **kwargs
    ) -> None:
        """Aggregate trajectories to construct training samples."""
        trajectories_with_pbar = tqdm(
            trajectories,
            desc="Aggregating trajectories",
            ncols=0,
            mininterval=1,
            disable=not show_pbar,
        )
        for trajectory in trajectories_with_pbar:
            self.aggregate_trajectory(trajectory, **kwargs)

    @abstractmethod
    def aggregate_trajectory(self, trajectory: Trajectory) -> None:
        """Aggregate transitions from a single trajectory to construct training samples."""

    @abstractmethod
    async def update(self) -> None:
        """Update the model based on the aggregated samples."""


class ChainedOptimizer(Optimizer):
    """An optimizer that runs a sequence of sub-optimizers in the order they are provided."""

    def __init__(self, *optimizers: Optimizer):
        self.optimizers = optimizers

    def aggregate(
        self, trajectories: Iterable[Trajectory], show_pbar: bool = False, **kwargs
    ) -> None:
        for optimizer in self.optimizers:
            optimizer.aggregate(trajectories, show_pbar=show_pbar, **kwargs)

    async def update(self) -> None:
        for optimizer in self.optimizers:
            await optimizer.update()



"""Test file for code examples extracted from the README.md file with a no-extra install."""

from dataclasses import dataclass, field
from datetime import datetime

import pytest

# =============================================================================
# Running an Agent on an Aviary Environment
# =============================================================================


@pytest.mark.asyncio
async def test_agent_on_aviary():
    """Test minimal example of running a language agent on an Aviary environment."""
    from aviary.core import DummyEnv

    from ldp.agent import SimpleAgent

    env = DummyEnv()
    agent = SimpleAgent()

    obs, tools = await env.reset()
    agent_state = await agent.init_state(tools=tools)

    # Run one step
    action, agent_state, _ = await agent.get_asv(agent_state, obs)
    obs, reward, done, truncated = await env.step(action.value)

    # Basic assertions
    assert action is not None
    assert agent_state is not None
    assert isinstance(reward, (int, float))
    assert isinstance(done, bool)
    assert isinstance(truncated, bool)


# =============================================================================
# SimpleAgent Definition
# =============================================================================


def test_simple_agent_definition():
    """Test defining LDP's SimpleAgent."""
    from ldp.agent import Agent
    from ldp.graph import LLMCallOp

    class AgentState:
        def __init__(self, messages, tools):
            self.messages = messages
            self.tools = tools

    class TestSimpleAgent(Agent):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.llm_call_op = LLMCallOp()

        async def init_state(self, tools):
            return AgentState([], tools)

        async def get_asv(self, agent_state, obs):
            # For testing purposes, create a mock action
            from aviary.core import ToolRequestMessage

            action = await Agent.wrap_action(ToolRequestMessage(tool_calls=[]))
            new_state = AgentState(
                messages=agent_state.messages + obs + [action], tools=agent_state.tools
            )
            return action, new_state, 0.0

    # Test that the class can be instantiated
    agent = TestSimpleAgent()
    assert agent is not None
    assert hasattr(agent, "llm_call_op")


# =============================================================================
# NoThinkAgent - Plain Python Agent
# =============================================================================


def test_no_think_agent():
    """Test minimal example of a deterministic Agent using plain Python."""
    from aviary.core import ToolCall, ToolRequestMessage

    from ldp.agent import Agent

    class NoThinkAgent(Agent):
        async def init_state(self, tools):
            return None

        async def get_asv(self, agent_state, obs):
            tool_call = ToolCall.from_name("specific_tool_call", arg1="foo")
            action = ToolRequestMessage(tool_calls=[tool_call])
            return await Agent.wrap_action(action), None, 0.0

    # Test that the class can be instantiated
    agent = NoThinkAgent()
    assert agent is not None


# =============================================================================
# Stochastic Computation Graph (SCG) - Basic Example
# =============================================================================


@pytest.mark.asyncio
async def test_basic_scg():
    """Test basic example of using stochastic computation graph."""
    from ldp.graph import FxnOp, compute_graph

    op_a = FxnOp(lambda x: 2 * x)

    async with compute_graph():
        op_result = await op_a(3)

    # Test that we got a result
    assert op_result is not None


# =============================================================================
# Complex SCG Example - Agent with Memory (Conceptual)
# =============================================================================


def test_memory_agent_structure():
    """Test the structure of the MemoryAgent class (conceptual only)."""

    class MemoryAgent:
        """Example agent that possesses memory using SCG."""

        def __init__(self):
            # These would be defined in the actual implementation
            self._query_factory_op = None
            self._memory_op = None
            self._format_memory_op = None
            self._prompt_op = None
            self._package_op = None
            self._config_op = None
            self._llm_call_op = None
            self.num_memories = 5
            self.memory_prompt = "Memory context:"

        def has_memory_components(self):
            """Check if the agent has all expected memory components."""
            return hasattr(self, "num_memories") and hasattr(self, "memory_prompt")

    # Test that the agent structure is correct
    memory_agent = MemoryAgent()
    assert memory_agent.has_memory_components()
    assert memory_agent.num_memories == 5
    assert memory_agent.memory_prompt == "Memory context:"


# =============================================================================
# Generic Support - Custom State Type
# =============================================================================


def test_custom_state_agent():
    """Test using Python generics with custom state type."""
    from ldp.agent import Agent

    @dataclass
    class MyComplexState:
        vector: list[float]
        timestamp: datetime = field(default_factory=datetime.now)

    class MyAgent(Agent[MyComplexState]):
        """Some agent who is now type checked to match the custom state."""

        async def init_state(self, tools):
            return MyComplexState(vector=[1.0, 2.0, 3.0])

        async def get_asv(self, agent_state, obs):
            from aviary.core import ToolRequestMessage

            action = await Agent.wrap_action(ToolRequestMessage(tool_calls=[]))
            return action, agent_state, 0.0

    # Test that the classes are defined correctly
    assert MyComplexState is not None
    assert MyAgent is not None

    # Create an instance of the state
    state_instance = MyComplexState(vector=[1.0, 2.0, 3.0])
    assert state_instance.vector == [1.0, 2.0, 3.0]
    assert isinstance(state_instance.timestamp, datetime)

    # Test that agent can be instantiated
    agent = MyAgent()
    assert agent is not None


# =============================================================================
# Rollout - Basic Example
# =============================================================================
@pytest.mark.asyncio
async def test_rollout():
    """Test basic example of running a rollout."""
    from aviary.core import DummyEnv

    from ldp.agent import SimpleAgent
    from ldp.alg import RolloutManager

    env = DummyEnv()
    agent = SimpleAgent()

    rollout_manager = RolloutManager(agent)
    traj, *_ = await rollout_manager.sample_trajectories(environments=[env])

    assert traj is not None



"""This module defines the Op class and its helper classes."""

from __future__ import annotations

import inspect
import itertools
import logging
import secrets
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Callable, Collection, Iterable, Iterator, Mapping, Sequence
from typing import TYPE_CHECKING, Any, ClassVar, Generic, TypeAlias, TypeVar
from uuid import UUID

if TYPE_CHECKING:
    import networkx as nx
    import tree
from pydantic import BaseModel, Field

from .op_utils import (
    CallID,
    _lazy_import_networkx,
    _lazy_import_tree,
    compute_graph,
    get_call_id,
    get_training_mode,
    op_call,
)

logger = logging.getLogger(__name__)


GradOutType: TypeAlias = (
    "tree.Structure | None"  # None means the gradient has terminated
)
GradInType: TypeAlias = tuple[Sequence[GradOutType], Mapping[str, GradOutType]]
BackwardsType: TypeAlias = Callable[
    # Call signature of Op.backward
    ["OpCtx", list, dict, "tree.Structure", CallID],
    GradInType,
]
TOutput_co = TypeVar("TOutput_co", covariant=True)


class OpResult(Generic[TOutput_co]):
    """Result of a forward pass, used in the compute graph."""

    def __init__(
        self, call_id: CallID | Any, op_name: str, op_class_name: str, value: TOutput_co
    ):
        """
        Initialize an OpResult instance.

        Args:
            call_id: The unique identifier for the call.
            op_name: Name of the Op instance (i.e. op.name) that produced this OpResult.
            op_class_name: Fully qualified name of the class of the Op that produced this OpResult.
            value: The output of the call.
        """
        self.call_id = CallID.model_validate(call_id)
        self.op_name = op_name
        self.op_class_name = op_class_name
        self.value = value

    def to_dict(self) -> dict[str, Any]:
        value_dump = (
            self.value.model_dump() if isinstance(self.value, BaseModel) else self.value
        )

        return {
            "call_id": self.call_id.model_dump(),
            "op_name": self.op_name,
            "op_class_name": self.op_class_name,
            "value": value_dump,
            "logprob": self.logprob,
        }

    @classmethod
    def from_dict(
        cls, t_output: type[TOutput_co], dump: dict[str, Any]
    ) -> OpResult[TOutput_co]:
        value = dump.pop("value")
        if issubclass(t_output, BaseModel):
            value = t_output.model_validate(value)

        # logprob was added to serialization for convenience,
        # but is not actually an OpResult argument, so we just discard it here
        dump.pop("logprob", None)
        return cls[t_output](**dump, value=value)  # type: ignore[index]

    def __hash__(self) -> int:
        return hash(self.call_id)

    def compute_grads(
        self,
        grad_output: tree.Structure | None = None,
        backward_fns: Mapping[str | type, BackwardsType] | None = None,
    ) -> None:
        """
        Compute the gradient of the backward graph in-place.

        This executes topological traversal.
        It is up to the Op to:
            (a) define the backward computation
            (b) store internal gradients for optimizer updates.
        """
        tree = _lazy_import_tree()

        # call ID -> [d op(x) / d x] for each op that consumes x[
        # due to interaction between ruff and mypy, we need type ignore
        grad_outputs: dict[CallID, list[tree.Structure]] = defaultdict(list)  # type: ignore[name-defined]

        # grad_outputs stores a list of output grads (corresponding to each consuming op call).
        # Since the root node is not consumed by any other node, we create a singleton list here.
        # If None was passed, set it to 0 so that we don't prune the compute graph here.
        grad_outputs[self.call_id] = [grad_output] if grad_output is not None else [0.0]

        # We will traverse the graph in reverse topological order
        for node in self.traverse():
            # get output gradients
            grad_output = grad_outputs[node.call_id]
            if not grad_output:
                # compute graph terminated
                continue
            # Make sure structure of grads match before summing
            try:
                [tree.assert_same_structure(grad_output[0], g) for g in grad_output[1:]]
            except ValueError as e:
                raise ValueError(
                    "Mismatched gradient structures in compute graph for at Op:"
                    f" {self.op_name}."
                ) from e
            aggregated_grad_output = tree.map_structure(lambda *x: sum(x), *grad_output)  # noqa: FURB111

            input_args, input_kwargs = node.inputs
            arg_grads, kwarg_grads = node._run_backward(
                input_args,
                input_kwargs,
                aggregated_grad_output,
                node._resolve_backward_impl(backward_fns),
            )

            for a, g in zip(input_args, arg_grads, strict=True):
                # Must have exact match between input_args and arg_grads
                # Only propagate gradients to input OpResults if grad is not None
                if g is not None and isinstance(a, OpResult):
                    grad_outputs[a.call_id].append(g)

            if kwarg_grads.keys() != input_kwargs.keys():
                raise ValueError(
                    "Mismatch between grads returned in Op.backward and its input"
                    f" kwargs. Expected {input_kwargs.keys()}, got"
                    f" {kwarg_grads.keys()}."
                )
            for k, a in input_kwargs.items():
                # input_kwargs.keys() may be a subset of kwarg_grads.keys() if defaults
                # are specified
                if (g := kwarg_grads[k]) is not None and isinstance(a, OpResult):
                    grad_outputs[a.call_id].append(g)

    def _resolve_backward_impl(
        self, backward_fns: Mapping[str | type, BackwardsType] | None
    ) -> BackwardsType:
        backward_fns = backward_fns or {}
        for key in (self.ctx.op_name, self.op_class_name, self.op_class):
            if key in backward_fns:
                return backward_fns[key]
        return self.op_class.backward

    def _run_backward(
        self,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: tree.Structure,
        backward_fn: BackwardsType,
    ) -> GradInType:
        self._update_ctx("grad_output", grad_output)
        unwrapped_input_args = [OpResult.unwrap_value(a) for a in input_args]
        unwrapped_input_kwargs = {
            k: OpResult.unwrap_value(v) for k, v in input_kwargs.items()
        }
        input_grads = backward_fn(
            self.ctx,
            unwrapped_input_args,
            unwrapped_input_kwargs,
            grad_output,
            self.call_id,
        )
        self._update_ctx("grad_input", input_grads)
        return input_grads

    @property
    def op_class(self) -> type[Op]:
        return _OP_CLASS_REGISTRY[self.op_class_name]

    @property
    def ctx(self) -> OpCtx:
        # This is a property to avoid serialization of the context. There are two reasons:
        # 1. Contexts have their own persist() mechanism for serialization
        # 2. We'd prefer contexts to be created via get_or_create(). Allowing for arbitrary
        #    deserialization makes it hard to enforce that.
        return OpCtx.get_or_create(self.op_name)

    @property
    def op(self) -> Op:
        return _OP_REGISTRY[self.op_name]

    def get_compute_graph(self, backward: bool = True) -> nx.DiGraph:
        """Construct a directed graph of the compute graph that led to this OpResult.

        Args:
            backward: If True (default), constructs the backwards graph in which outputs
                point to inputs. If False, constructs the forward call graph.
                For most cases (e.g. backprop), backward=True is desirable.

        Returns:
            A digraph in which nodes are OpResults.
        """

        def add_edges(graph: nx.DiGraph, node: OpResult) -> None:
            """Recursively add edges to the input graph."""
            input_args, input_kwargs = node.inputs
            for x in itertools.chain(input_args, input_kwargs.values()):
                if isinstance(x, OpResult):
                    edge = (node, x) if backward else (x, node)
                    graph.add_edge(*edge)
                    add_edges(graph, x)

        graph = _lazy_import_networkx().DiGraph()
        graph.add_node(self)
        add_edges(graph, self)

        return graph

    def get_upstream_results(self, op: str | Op) -> Iterator[OpResult]:
        """Get all OpResults upstream of this node that were produced by the given Op.

        Args:
            op: Will return all upstream nodes that were produced by this Op. Can provide
                 an instance or Op name.
        """
        if isinstance(op, Op):
            op = op.name

        return self.traverse(filter_fn=lambda node: node.op_name == op)

    def traverse(
        self,
        topological_order: bool = True,
        filter_fn: Callable[[OpResult], bool] = lambda _: True,
    ) -> Iterator[OpResult]:
        """Traverse the compute graph that led to this OpResult.

        Args:
            topological_order: If True, traverse the backwards graph in topological
                order. This requires having the whole graph in memory. If False,
                traverse the backwards graph in depth-first order. This can be done
                lazily and is useful if we are trying to hydrate the graph node-by-node.
                Most user-facing cases can leave this as True. Defaults to True.
            filter_fn: Will only yield nodes that pass this filter function. Note that
                nodes that fail will still be traversed.

        Yields:
            An iterator over the nodes of this graph.
        """
        if topological_order:
            G = self.get_compute_graph()
            for node in _lazy_import_networkx().topological_sort(G):
                if filter_fn(node):
                    yield node

        else:
            # If not topological order, do a recursive depth-first traversal.
            # Note that, when traversing a node, its children do not need to be available
            # yet. This allows us to lazily load nodes when hydrating from a ctx backend.
            if filter_fn(self):
                yield self
            input_args, input_kwargs = self.inputs
            for a in itertools.chain(input_args, input_kwargs.values()):
                if isinstance(a, OpResult):
                    # Recursively apply depth-first traversal on each node
                    yield from a.traverse(topological_order=False)

    @property
    def inputs(self) -> tuple[list[ResultOrValue], dict[str, ResultOrValue]]:
        return self._get_from_ctx("input")

    @property
    def logprob(self) -> float | None:
        return self._get_from_ctx("logprob", default=None)

    @property
    def grad(self) -> tree.Structure | None:
        """Returns `d ln(P_{compute_graph}) / d self` or None if gradients have not been computed."""
        return self._get_from_ctx("grad_output")

    @property
    def run_id(self) -> UUID:
        return self.call_id.run_id

    @staticmethod
    def unwrap_value(result: ResultOrValue[TOutput_co]) -> TOutput_co:
        if isinstance(result, OpResult):
            return result.value
        return result

    def __repr__(self) -> str:
        return (
            f"OpResult(op={self.op_class_name}:{self.op_name}, "
            f"call_id={self.call_id}, value={self.value!r})"
        )

    def __str__(self) -> str:
        return str(self.value)

    def _get_from_ctx(self, key: str, **kwargs):
        if self.call_id is None:
            raise ValueError(
                "Attempting to access context but compute graph "
                "is not available for this OpResult."
            )

        return self.ctx.get(call_id=self.call_id, key=key, **kwargs)

    def _update_ctx(self, key: str, value: Any):
        if self.call_id is None:
            raise RuntimeError(
                "Attempting to update context but compute graph "
                "is not available for this OpResult."
            )

        self.ctx.update(call_id=self.call_id, key=key, value=value)


ResultOrValue: TypeAlias = OpResult[TOutput_co] | TOutput_co

# Sentinel value for get() default
NOT_FOUND = object()


class OpCtx(BaseModel):
    # A global registry of contexts. We'd prefer to use an existing context
    # for an Op if it already has been created. Also useful for persist_all()
    # NOTE: see the comment on _OP_REGISTRY below on a potential memory leak.
    _CTX_REGISTRY: ClassVar[dict[str, OpCtx]] = {}

    op_name: str

    data: dict = Field(
        default_factory=lambda: defaultdict(dict),
        exclude=True,
        description=(
            "Maps run_id -> (fwd_id, key) -> value. "
            "data is excluded from model_dump() etc because we do "
            "not use Pydantic to persist context information. That "
            "should be done via the DB backend instead. OpCtx will "
            "serialize op_name, which is enough to rehydrate "
            "from the DB."
        ),
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._CTX_REGISTRY[self.op_name] = self

    @classmethod
    def get_or_create(cls, op_name: str) -> OpCtx:
        """Return an OpCtx corresponding to the Op with the given name."""
        try:
            return cls._CTX_REGISTRY[op_name]  # Get
        except KeyError:
            return cls(op_name=op_name)  # Create

    @classmethod
    def clear_contexts(cls, op_names: Iterable[str] | None = None) -> None:
        """Clear the data in all contexts. If op_names is provided, only clear those contexts."""
        if op_names is None:
            op_names = cls._CTX_REGISTRY.keys()
        for op_name in op_names:
            try:
                cls._CTX_REGISTRY[op_name].data.clear()
            except KeyError:
                logger.warning(f"Op with name={op_name} not found in context registry.")

    def get(self, call_id: CallID, key: str, default: Any = NOT_FOUND) -> Any:
        """Get an attribute with an optional default, emulating dict.get."""
        value = self.data.get(call_id.run_id, {}).get((call_id.fwd_id, key), default)
        if value is NOT_FOUND:
            raise KeyError(f"call_id={call_id}, key='{key}' not found in context")
        return value

    def update(self, call_id: CallID, key: str, value: Any):
        self.data[call_id.run_id][call_id.fwd_id, key] = value

    def get_input_grads(self, call_id: CallID) -> GradInType:
        # TODO: this function name is confusing. Let's deprecate it. We only use it
        # in tests as far as I can tell.
        try:
            return self.get(call_id, "grad_input")
        except KeyError as exc:
            raise ValueError(
                f"No gradients have been computed for call_id={call_id}."
            ) from exc


def resolve_fully_qualified_name(cls: type) -> str:
    return f"{cls.__module__}.{cls.__name__}"


# A global registry of Op classes, so we can look up backward() implementations
# without needing an instantiated Op.
_OP_CLASS_REGISTRY: dict[str, type[Op]] = {}

# A global registry of Op instances, so that OpResults can trace back their provenance
# TODO: this can leak memory if we are frequently deleting Ops. We should add a custom
# garbage collection hook to remove Ops from the registry once the reference count
# drops to 1 (i.e. just the registry).
_OP_REGISTRY: dict[str, Op] = {}


class Op(ABC, Generic[TOutput_co]):
    """
    An operation that is 'differentiable' and can be used in an optimizer.

    Think torch.autograd.Function that can also be applied to non-differentiable
    operations like prompt template formatting or Python function calls.

    These form a forward computation graph when composed with other Ops via
    __call__. In training mode, this graph is constructed dynamically.
    """

    # Name is not guaranteed to be unique. Reasons:
    # 1. We definitely don't want it to be unique when recreating a compute graph
    #    for training on previously-collected data. In that case, we want the Op's
    #    name to be the same as it was before, to match up contexts/OpResults
    # 2. Uniqueness could make some DB lookups faster, but I don't think we run the
    #    risk of OpCtx clobbers as long as call_id (which is guaranteed to be unique)
    #    is always used as part of the key.
    name: str
    ctx: OpCtx
    _fwd_args: list[inspect.Parameter]

    def clear_ctx(self) -> None:
        self.ctx.data.clear()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        _OP_CLASS_REGISTRY[resolve_fully_qualified_name(cls)] = cls

    def __new__(cls, *args, **kwargs):  # noqa: ARG004
        instance = super().__new__(cls)

        # Needs to be overridden by caller if this Op is to have
        # an identifiable name in the compute graph. c.f. Agent.__init_subclass__
        # for an example of how to do this.
        instance.set_name(cls._make_unique_default_name())

        # Set an attribute to help us map positional forward arguments to parameter
        # names, for the backward pass. We do this on the instance and not cls b/c
        # some instancees may override (e.g FxnOp).
        fwd_sig = inspect.signature(instance.forward)
        instance._fwd_args = list(fwd_sig.parameters.values())

        return instance

    @classmethod
    def _make_unique_default_name(cls) -> str:
        # 6 bytes results in string of size 12
        return f"{cls.__name__}-{secrets.token_hex(6)}"

    def set_name(self, name: str) -> None:
        if _OP_REGISTRY.get(getattr(self, "name", "")) is self:
            # de-register before setting the new name
            del _OP_REGISTRY[self.name]

        self.name = name
        self.ctx = OpCtx.get_or_create(name)
        _OP_REGISTRY[name] = self

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} (name={self.name}, id={id(self)})"

    @abstractmethod
    async def forward(self, *args, **kwargs) -> TOutput_co:
        """
        Forward pass of the Op. Must accept call_id as an argument.

        Returns:
            Depending on this Op's purpose, the return may be considered an action
                (e.g. a tool call) or it may not (e.g. a loss calculation).
        """

    @classmethod
    @abstractmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs,
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        """
        Backward pass of the Op.

        Args:
            ctx: Context that was used during the forward pass.
            input_args: Variable-length input arguments passed to forward, i.e.
                via *args.
            input_kwargs: All other arguments passed to forward pass.
            grad_output: A list of backpropagated gradients from each consumer
                of the output of the forward pass. It is up to the implementation
                to decide how to aggregate these gradients (e.g. in most cases summing).
            call_id: Call ID of the forward pass.

        Returns:
            grad_input: `d log(p) / d input` for each input to the forward pass.
                It should include gradients for all input positional and keyword
                arguments. Set to None for gradients that should terminate.
        """

    def get_call_ids(self, run_ids: Collection[UUID] | None = None) -> set[CallID]:
        ctx = self.ctx
        if run_ids is None:
            run_ids = ctx.data.keys()

        # de-duplicate before constructing CallIDs
        ids = {(run_id, fwd_id) for run_id in run_ids for fwd_id, _ in ctx.data[run_id]}
        return set(itertools.starmap(CallID, ids))

    # This compute_graph() decoration will do nothing if we are already inside a compute graph.
    # We add it here in case we are calling a bare op(), in which case we want a graph
    # with a single node.
    @compute_graph()
    @op_call()
    async def __call__(self, *args, **kwargs) -> OpResult[TOutput_co]:
        call_id = get_call_id()

        if not all(
            arg.call_id.run_id == call_id.run_id
            for arg in itertools.chain(args, kwargs.values())
            if isinstance(arg, OpResult)
        ):
            raise RuntimeError(
                "All args and kwargs must have the same run_id as the call_id's run_id."
                " Consider using @compute_graph() decorator to ensure this."
            )

        # we're over-saving here - can explore later if memory usage is high
        # unpack the args and kwargs from the result holders
        unpacked_args = [(a.value if isinstance(a, OpResult) else a) for a in args]
        unpacked_kwargs = {
            k: v.value if isinstance(v, OpResult) else v for k, v in kwargs.items()
        }

        if get_training_mode():
            # If training, save the inputs for the backward pass
            # Map positional arguments to keyword arguments to make backward pass easier
            for i_arg, (arg, param) in enumerate(
                # strict=False b/c not all params in _fwd_args will be in args (i.e. defaults and **kwargs)
                zip(args, self._fwd_args, strict=False)
            ):
                # Don't need to check for too many args or collisions with kwargs, since forward()
                # will raise an exception anyway
                if param.kind == inspect.Parameter.VAR_POSITIONAL:
                    # *args, so scoop up the rest of the arg tuple.
                    var_args = list(args[i_arg:])
                    break

                # Normal positional arg
                kwargs[param.name] = arg
            else:
                var_args = []  # there were no *args if we got here

            self.ctx.update(call_id, "input", (var_args, kwargs))

        # actually call forward pass with unpacked args and kwargs
        result = await self.forward(*unpacked_args, **unpacked_kwargs)
        t_output: type[TOutput_co] = type(result)

        # Now package up my result so it can be consumed by other calls.
        # Explicitly specify t_output. OpResult[TOutput] returns a generic object
        op_result = OpResult[t_output](  # type: ignore[valid-type]
            value=result,
            call_id=call_id,
            op_name=self.name,
            op_class_name=resolve_fully_qualified_name(type(self)),
        )

        if get_training_mode():
            self.ctx.update(call_id, "output", op_result)

        return op_result

    def get_input_grads(self, call_id: CallID) -> GradInType:
        return self.ctx.get_input_grads(call_id)



import os
import random
from collections.abc import AsyncIterator
from typing import Any

import httpx_aiohttp
import litellm.llms.custom_httpx.aiohttp_transport
import numpy as np
import pytest
import torch
import vcr.stubs.httpx_stubs
from aviary.core import DummyEnv
from lmi import configure_llm_logs
from lmi.utils import update_litellm_max_callbacks

from ldp.nn.handlers.transformer_handler import ExecutionMode, ParallelModeConfig

from . import CASSETTES_DIR

IN_GITHUB_ACTIONS: bool = os.getenv("GITHUB_ACTIONS") == "true"


@pytest.fixture(name="dummy_env")
def fixture_dummy_env() -> DummyEnv:
    return DummyEnv()


@pytest.fixture(scope="session", autouse=True)
def _fixture_set_up_environment() -> None:
    configure_llm_logs()


@pytest.fixture(autouse=True, scope="session")
def _defeat_litellm_callbacks() -> None:
    update_litellm_max_callbacks()


def set_seed(seed: int | None) -> None:
    if seed is None:
        return

    random.seed(seed)
    np.random.seed(seed)  # noqa: NPY002
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)


@pytest.fixture(name="seed_zero")
def fixture_seed_zero() -> None:
    """Set a 0 seed to minimize the chances of test flakiness."""
    set_seed(0)


OPENAI_API_KEY_HEADER = "authorization"
ANTHROPIC_API_KEY_HEADER = "x-api-key"
# SEE: https://github.com/kevin1024/vcrpy/blob/v6.0.1/vcr/config.py#L43
VCR_DEFAULT_MATCH_ON = "method", "scheme", "host", "port", "path", "query"


@pytest.fixture(scope="session", name="vcr_config")
def fixture_vcr_config() -> dict[str, Any]:
    return {
        "filter_headers": [OPENAI_API_KEY_HEADER, ANTHROPIC_API_KEY_HEADER, "cookie"],
        "record_mode": "once",
        "match_on": ["method", "host", "path", "query"],
        "allow_playback_repeats": True,
        "cassette_library_dir": str(CASSETTES_DIR),
    }


ENABLED = {"true", "1", "yes"}
TEST_GPUS: bool = os.getenv("TEST_GPUS", "").lower() in ENABLED
TEST_SLURM: bool = os.getenv("TEST_SLURM", "").lower() in ENABLED

PARALLEL_MODE_CONFIGS = [
    pytest.param(None, id="cpu-only"),
    pytest.param(
        ParallelModeConfig(num_workers=2, num_cpus_per_worker=1),
        id="two-gpu",
        marks=pytest.mark.skipif(not TEST_GPUS, reason="Requires GPUs"),
    ),
    pytest.param(
        ParallelModeConfig(num_workers=2, num_cpus_per_worker=1, offload_cpu=True),
        id="two-gpu-offload",
        marks=pytest.mark.skipif(not TEST_GPUS, reason="Requires GPUs"),
    ),
    pytest.param(
        ParallelModeConfig(
            num_workers=2,
            num_cpus_per_worker=1,
            offload_cpu=True,
            activation_checkpointing=True,
            cpu_ram_efficient_loading=True,
        ),
        id="two-gpu-all-enabled",
        marks=pytest.mark.skipif(not TEST_GPUS, reason="Requires GPUs"),
    ),
    pytest.param(
        ParallelModeConfig(
            num_workers=2,
            num_cpus_per_worker=1,
            execution_mode=ExecutionMode.SLURM_CLUSTER,
        ),
        id="two-gpus-slurm",
        marks=pytest.mark.skipif(
            not TEST_GPUS or not TEST_SLURM, reason="Requires GPUs and SLURM"
        ),
    ),
    pytest.param(
        ParallelModeConfig(
            num_workers=2,
            num_cpus_per_worker=1,
            execution_mode=ExecutionMode.SLURM_CLUSTER,
            offload_cpu=True,
        ),
        id="two-gpus-slurm-offload",
        marks=pytest.mark.skipif(
            not TEST_GPUS or not TEST_SLURM, reason="Requires GPUs and SLURM"
        ),
    ),
]


class PreReadCompatibleAiohttpResponseStream(
    httpx_aiohttp.transport.AiohttpResponseStream
):
    """aiohttp-backed response stream that works if the response was pre-read."""

    async def __aiter__(self) -> AsyncIterator[bytes]:
        with httpx_aiohttp.transport.map_aiohttp_exceptions():
            if self._aiohttp_response._body is not None:
                # Happens if some intermediary called `await _aiohttp_response.read()`
                # TODO: take into account chunk size
                yield self._aiohttp_response._body
            else:
                async for chunk in self._aiohttp_response.content.iter_chunked(
                    self.CHUNK_SIZE
                ):
                    yield chunk


async def _async_vcr_send(cassette, real_send, *args, **kwargs):  # noqa: ARG001
    """VCR send that only sends, not possibly recording or playing back responses."""
    return await real_send(*args, **kwargs)


# Permanently patch the original response stream,
# to work around https://github.com/karpetrosyan/httpx-aiohttp/issues/23
# and https://github.com/BerriAI/litellm/issues/11724
httpx_aiohttp.transport.AiohttpResponseStream = (  # type: ignore[misc]
    litellm.llms.custom_httpx.aiohttp_transport.AiohttpResponseStream  # type: ignore[misc]
) = PreReadCompatibleAiohttpResponseStream  # type: ignore[assignment]

# Permanently patch vcrpy's async VCR recording functionality,
# to work around https://github.com/kevin1024/vcrpy/issues/944
vcr.stubs.httpx_stubs._async_vcr_send = _async_vcr_send



from .algorithms import (
    bulk_evaluate_consensus,
    compute_pass_at_k,
    evaluate_consensus,
    to_network,
)
from .beam_search import Beam, BeamSearchRollout
from .callbacks import (
    Callback,
    ClearContextCallback,
    ClearOptimizerBuffersCallback,
    ComputeTrajectoryMetricsMixin,
    LoggingCallback,
    MeanMetricsCallback,
    RolloutDebugDumpCallback,
    StoreEnvironmentsCallback,
    StoreTrajectoriesCallback,
    TrajectoryFileCallback,
    TrajectoryMetricsCallback,
    WandBLoggingCallback,
)
from .rollout import RolloutManager
from .runners import (
    Evaluator,
    EvaluatorConfig,
    OfflineTrainer,
    OfflineTrainerConfig,
    OnlineTrainer,
    OnlineTrainerConfig,
)
from .tree_search import TEnvCloneFn, TreeSearchRollout

__all__ = [
    "Beam",
    "BeamSearchRollout",
    "Callback",
    "ClearContextCallback",
    "ClearOptimizerBuffersCallback",
    "ComputeTrajectoryMetricsMixin",
    "Evaluator",
    "EvaluatorConfig",
    "LoggingCallback",
    "MeanMetricsCallback",
    "OfflineTrainer",
    "OfflineTrainerConfig",
    "OnlineTrainer",
    "OnlineTrainerConfig",
    "RolloutDebugDumpCallback",
    "RolloutManager",
    "StoreEnvironmentsCallback",
    "StoreTrajectoriesCallback",
    "TEnvCloneFn",
    "TrajectoryFileCallback",
    "TrajectoryMetricsCallback",
    "TreeSearchRollout",
    "WandBLoggingCallback",
    "bulk_evaluate_consensus",
    "compute_pass_at_k",
    "evaluate_consensus",
    "to_network",
]



import asyncio
import uuid
from collections.abc import Awaitable, Callable, Sequence
from contextlib import suppress
from copy import deepcopy
from typing import NamedTuple, cast

from aviary.core import Environment

from ldp.agent.agent import Agent, TAgentState
from ldp.data_structures import Trajectory, Transition

from .callbacks import Callback
from .rollout import AgentError, EnvError, TEnv, reraise_exc_as


class Beam(NamedTuple):
    # An ongoing beam contains two things: the trajectory up to now
    # and the environment that the last action was sampled from. We
    # need both to continue sampling the next step.
    traj: Trajectory
    env: Environment


class BeamSearchRollout:
    def __init__(
        self,
        agent: Agent,
        beam_width: int,
        samples_per_beam: int,
        env_clone_fn: Callable[[TEnv], Awaitable[TEnv]],
        agent_clone_fn: Callable[[TAgentState], TAgentState],
        scoring_fn: Callable[[Trajectory], Awaitable[float]],
        replay_actions_on_clone: bool = False,
        callbacks: Sequence[Callback] | None = None,
        catch_agent_failures: bool = True,
        catch_env_failures: bool = True,
        verbose: bool = False,
    ):
        self.agent = agent

        self.catch_agent_failures = catch_agent_failures
        self.catch_env_failures = catch_env_failures

        self.verbose = verbose

        self.traj_buffer: dict[str, Trajectory] = {}
        self.search_buffer: dict[str, list[Trajectory]] = {}

        self.beam_width = beam_width
        self.samples_per_beam = samples_per_beam

        self.env_clone_fn = env_clone_fn
        self.agent_clone_fn = agent_clone_fn
        self.scoring_fn = scoring_fn
        self.replay_actions_on_clone = replay_actions_on_clone

        self.callbacks = callbacks or []

    async def sample_trajectories(
        self,
        environments: Sequence[Environment],
        max_steps: int | None = None,
    ) -> list[Trajectory]:
        self.traj_buffer.clear()
        traj_ids = [uuid.uuid4().hex for _ in environments]

        tasks = [
            self._rollout(traj_id, env, max_steps)
            for traj_id, env in zip(traj_ids, environments, strict=True)
        ]
        await asyncio.gather(*tasks)

        return [self.traj_buffer[traj_id] for traj_id in traj_ids]

    async def _rollout(
        self, traj_id: str, env: Environment, max_steps: int | None
    ) -> None:
        with suppress(AgentError, EnvError):
            # for samples_per_beam==1. we want to ensemble and pick the highest-scoring one
            n_seeds = 1 if self.samples_per_beam > 1 else self.beam_width

            done_beams: list[Beam] = []
            beams = [
                Beam(
                    traj=await Trajectory.from_env(env, traj_id=f"{traj_id}:{i}"),
                    env=env,
                )
                for i in range(n_seeds)
            ]
            # will be replaced if rollout is successful
            self.traj_buffer[traj_id] = beams[0].traj
            self.search_buffer[traj_id] = []

            await asyncio.gather(*[
                c.before_rollout(traj_id, env) for c in self.callbacks
            ])

            with reraise_exc_as(EnvError, self.catch_env_failures):
                init_obs, tools = await env.reset()
            await asyncio.gather(*[
                c.after_env_reset(traj_id, init_obs, tools) for c in self.callbacks
            ])

            with reraise_exc_as(AgentError, self.catch_agent_failures):
                seed_agent_states = await asyncio.gather(
                    *(self.agent.init_state(tools) for _ in range(n_seeds))
                )
            # TODO: implement after_agent_init_state callback

            while len(done_beams) < self.beam_width and beams:
                new_beams = []
                for beam, seed_agent_state in zip(
                    beams, seed_agent_states, strict=True
                ):
                    for i_sample in range(self.samples_per_beam):
                        new_env = await self._clone_env(beam)
                        if new_env is None:
                            continue

                        agent_state = self.agent_clone_fn(seed_agent_state)
                        obs = (
                            beam.traj.steps[-1].next_observation
                            if beam.traj.steps
                            else init_obs.copy()
                        )

                        await asyncio.gather(*[
                            callback.before_transition(
                                traj_id, self.agent, env, agent_state, obs
                            )
                            for callback in self.callbacks
                        ])

                        with reraise_exc_as(AgentError, self.catch_agent_failures):
                            (
                                action,
                                next_agent_state,
                                vhat,
                            ) = await self.agent.get_asv(agent_state, obs)
                        await asyncio.gather(*[
                            callback.after_agent_get_asv(
                                traj_id, action, next_agent_state, vhat
                            )
                            for callback in self.callbacks
                        ])

                        with reraise_exc_as(EnvError, self.catch_env_failures):
                            next_obs, reward, done, trunc = await new_env.step(
                                action.value
                            )
                        await asyncio.gather(*[
                            callback.after_env_step(
                                traj_id, next_obs, reward, done, trunc
                            )
                            for callback in self.callbacks
                        ])

                        step = Transition(
                            timestep=len(beam.traj.steps),
                            agent_state=agent_state,
                            next_agent_state=next_agent_state,
                            observation=obs,
                            next_observation=next_obs,
                            action=action,
                            reward=reward,
                            done=done,
                            truncated=trunc,
                            value=0.0,  # will be filled in
                        )
                        await asyncio.gather(*[
                            callback.after_transition(traj_id, self.agent, env, step)
                            for callback in self.callbacks
                        ])

                        new_beam = Beam(
                            traj=Trajectory(
                                traj_id=cast(str, beam.traj.traj_id) + f":{i_sample}",
                                steps=[*beam.traj.steps, step],
                                metadata=deepcopy(beam.traj.metadata),
                            ),
                            env=new_env,
                        )
                        step.value = await self.scoring_fn(new_beam.traj)
                        self.search_buffer[traj_id].append(new_beam.traj)

                        if (
                            not new_beam.traj.done
                            and max_steps is not None
                            and len(new_beam.traj.steps) >= max_steps
                        ):
                            last_step = new_beam.traj.steps[-1]
                            last_step.done = last_step.truncated = True

                        if new_beam.traj.done:
                            done_beams.append(new_beam)
                        else:
                            new_beams.append(new_beam)

                new_beams.sort(key=lambda b: b.traj.steps[-1].value, reverse=True)
                beams, discarded = (
                    new_beams[: self.beam_width],
                    new_beams[self.beam_width :],
                )
                seed_agent_states = [b.traj.steps[-1].next_agent_state for b in beams]
                await asyncio.gather(*[d.env.close() for d in discarded])

            await asyncio.gather(*[b.env.close() for b in beams])

            self.traj_buffer[traj_id] = max(
                done_beams,
                key=lambda b: (b.traj.steps[-1].truncated, b.traj.steps[-1].value),
            ).traj

    async def _clone_env(self, beam: Beam) -> Environment | None:
        try:
            with reraise_exc_as(EnvError, self.catch_env_failures):
                # I'm not sure how to type hint this
                env = await self.env_clone_fn(beam.env)  # type: ignore[arg-type]
                if self.replay_actions_on_clone:
                    # Some envs can't be cloned, so instead replay.
                    # We rely on env_clone_fn to properly reset the env if needed.
                    # We assume a deterministic env, so the return values are discarded.
                    for step in beam.traj.steps:
                        if step.action is not None:
                            _ = await env.step(step.action.value)
                return env
        except EnvError:
            return None



from __future__ import annotations

from collections.abc import Awaitable, Callable
from itertools import chain
from typing import Any, Self, cast

from aviary.core import Message, Tool, ToolRequestMessage, ToolResponseMessage
from aviary.message import EnvStateMessage
from lmi import CommonLLMNames
from pydantic import BaseModel, ConfigDict, Field

from ldp.graph import ConfigOp, FxnOp, LLMCallOp, OpResult, compute_graph
from ldp.llms import prepend_sys
from ldp.utils import split_message_transitions

from . import DEFAULT_LLM_COMPLETION_TIMEOUT
from .agent import Agent


class HiddenEnvStateMessage(EnvStateMessage):
    content: str = "[Previous environment state - hidden]"


def hide_action_content(msg: ToolRequestMessage) -> ToolRequestMessage:
    return msg.model_copy(update={"content": None})


class SimpleAgentState(BaseModel):
    """Simple bucket for an Agent to access tools and store messages."""

    model_config = ConfigDict(extra="forbid")

    tools: list[Tool] = Field(default_factory=list)
    messages: list[ToolRequestMessage | ToolResponseMessage | Message] = Field(
        default_factory=list
    )
    hide_old_env_states: bool = Field(
        default=False,
        description="Whether to hide old EnvStateMessages.",
    )
    hide_old_action_content: bool = Field(
        default=False,
        description="If True, will hide the content of old ToolRequestMessages.",
    )
    sliding_window: int | None = Field(
        default=None,
        description="Number of previous trajectory transitions to keep. None means all previous transitions.",
    )

    def get_next_state(
        self,
        obs: list[Message] | None = None,
        tools: list[Tool] | None = None,
        hide_old_env_states: bool | None = None,
        **kwargs,
    ) -> Self:
        """
        Get the next agent state without mutating the optional prior state.

        Do not mutate self here, just read from it.

        Args:
            obs: Optional observation messages to use in creating the next state.
            tools: Optional list of tools available to the agent. If unspecified, these
                should be pulled from the prior_state.
            hide_old_env_states: Optional override of self.hide_old_env_states.
                TODO: do we still need this override?
            kwargs: Additional keyword arguments to pass to this class's constructor.

        Returns:
            The next agent state (which is not an in-place change to self).
        """
        old_messages = self.messages
        hide_old_env_states = (
            hide_old_env_states
            if hide_old_env_states is not None
            else self.hide_old_env_states
        )
        if self.sliding_window is not None and self.sliding_window > 0 and old_messages:
            hide_message = "[Previous messages - hidden]"
            msg_blocks = split_message_transitions(old_messages)

            # Do not duplicate hide_message if it's already in the first block
            if msg_blocks[0][-1].content != hide_message:
                msg_blocks[0].append(Message(content=hide_message))

            old_messages = (
                msg_blocks[0]  # keep system messages + user message + hide_message
                + list(chain.from_iterable(msg_blocks[1:][-self.sliding_window :]))
            )

        if hide_old_env_states:
            old_messages = [
                HiddenEnvStateMessage() if isinstance(m, EnvStateMessage) else m
                for m in old_messages
            ]
        if self.hide_old_action_content:
            old_messages = [
                hide_action_content(m) if isinstance(m, ToolRequestMessage) else m
                for m in old_messages
            ]

        return type(self)(
            tools=tools if tools is not None else self.tools,
            messages=old_messages + (obs or []),
            hide_old_env_states=hide_old_env_states,
            hide_old_action_content=self.hide_old_action_content,
            sliding_window=self.sliding_window,
            **kwargs,
        )


class SimpleAgent(BaseModel, Agent[SimpleAgentState]):
    """Simple agent that can pick and invoke tools with a language model."""

    # Freeze to ensure the only mutation happens in either the agent state (which is
    # passed around) or in the internal Ops
    model_config = ConfigDict(frozen=True)

    llm_model: dict[str, Any] = Field(
        default={
            "name": CommonLLMNames.GPT_4O.value,
            "temperature": 0.1,
            "timeout": DEFAULT_LLM_COMPLETION_TIMEOUT,
        },
        description="Starting configuration for the LLM model. Trainable.",
    )
    sys_prompt: str | None = Field(
        default=None,
        description=(
            "Opt-in system prompt. If one is passed, the system prompt is not set up to"
            " be trainable, because this class is meant to be quite simple as far as"
            " possible hyperparameters."
        ),
    )

    hide_old_env_states: bool = Field(
        default=False,
        description="See SimpleAgentState.hide_old_env_states.",
    )

    hide_old_action_content: bool = Field(
        default=False,
        description="See SimpleAgentState.hide_old_action_content.",
    )

    sliding_window: int | None = Field(
        default=None,
        description="See SimpleAgentState.sliding_window.",
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._config_op = ConfigOp[dict](config=self.llm_model)
        self._llm_call_op = LLMCallOp()

    async def init_state(self, tools: list[Tool]) -> SimpleAgentState:
        return SimpleAgentState(
            tools=tools,
            hide_old_env_states=self.hide_old_env_states,
            hide_old_action_content=self.hide_old_action_content,
            sliding_window=self.sliding_window,
        )

    @compute_graph()
    async def get_asv(
        self, agent_state: SimpleAgentState, obs: list[Message]
    ) -> tuple[OpResult[ToolRequestMessage], SimpleAgentState, float]:
        next_state = agent_state.get_next_state(obs)

        messages = (
            prepend_sys(next_state.messages, sys_content=self.sys_prompt)
            if self.sys_prompt is not None
            else next_state.messages
        )
        result = cast(
            "OpResult[ToolRequestMessage]",
            await self._llm_call_op(
                await self._config_op(), msgs=messages, tools=next_state.tools
            ),
        )
        next_state.messages = [*next_state.messages, result.value]
        return result, next_state, 0.0


class NoToolsSimpleAgent(SimpleAgent):
    """Agent whose action is parsed out of a LLM prompt without tool schemae."""

    def __init__(
        self,
        cast_tool_request: (
            Callable[[Message], ToolRequestMessage]
            | Callable[[Message], Awaitable[ToolRequestMessage]]
        ),
        **kwargs,
    ):
        """Initialize.

        Args:
            cast_tool_request: Function that is given a plain text message and produces
                a tool request message.
            **kwargs: Keyword arguments to pass to SimpleAgent's constructor.
        """
        super().__init__(**kwargs)
        self._cast_tool_request_op = FxnOp[ToolRequestMessage](cast_tool_request)

    @compute_graph()
    async def get_asv(
        self, agent_state: SimpleAgentState, obs: list[Message]
    ) -> tuple[OpResult[ToolRequestMessage], SimpleAgentState, float]:
        next_state = agent_state.get_next_state(obs)

        messages = (
            prepend_sys(next_state.messages, sys_content=self.sys_prompt)
            if self.sys_prompt is not None
            else next_state.messages
        )
        result = await self._cast_tool_request_op(
            # NOTE: this call has no tools specified to the LLM
            await self._llm_call_op(await self._config_op(), msgs=messages)
        )
        next_state.messages = [*next_state.messages, result.value]
        return result, next_state, 0.0



"""This module defines the MemoryAgent class, which extends a base agent model with memory.

capabilities. The MemoryAgent can pick and invoke tools based on the stored and retrieved
memories, formatted using specified prompts. A memory is typically a set of previous trajectories
"""

from collections.abc import Awaitable, Callable, Iterable
from typing import ClassVar, cast

from aviary.core import Message, ToolRequestMessage
from pydantic import ConfigDict, Field

from ldp.graph import (
    FxnOp,
    Memory,
    MemoryModel,
    MemoryOp,
    OpResult,
    PromptOp,
    compute_graph,
)
from ldp.llms.prompts import indent_xml

from .simple_agent import SimpleAgent, SimpleAgentState


async def _default_query_factory(messages: Iterable[Message]) -> str:  # noqa: RUF029
    return "\n\n".join([str(m) for m in messages if m.role != "system"])


class MemoryAgent(SimpleAgent):
    """
    Simple agent that can pick and invoke tools with memory.

    NOTE: the MemoryAgent does not maintain an explicit value estimate,
    it simply supplies previous trajectories via the prompt.
    As such, the value estimate vhat will always be zero.
    """

    # Working around https://github.com/pydantic/pydantic/issues/10551
    default_query_factory: ClassVar[Callable[[Iterable[Message]], Awaitable[str]]] = (
        _default_query_factory
    )

    prompt: str = Field(
        default=(
            "<episode-memories>\n<description>"
            "\nThese are relevant memories from previous attempts at similar tasks,"
            " along with the action taken"
            " and the discounted cumulative reward from that action."
            " A negative reward is failure, a positive reward is success."
            "\n</description>{memories}</episode-memories>"
            "\n\nConsidering the memories, choose the next action."
        ),
        description="Prompt that includes the memories.",
    )
    query_factory: Callable[[Iterable[Message]], Awaitable[str]] = Field(
        default=default_query_factory,
        description=(
            "Async function to generate a Memory query string from messages. It's async"
            " so this can involve an LLM completion if desired."
        ),
        exclude=True,
    )
    memory_prompt: str = Field(
        default="<memory><obs>{input}</obs><action>{output}</action><reward>{value}</reward></memory>",
        description=(
            "Prompt for formatting an individual memory. "
            "Use XML instead of JSON to avoid potential escaping issues."
        ),
    )
    num_memories: int = Field(
        default=MemoryModel.DEFAULT_MEMORY_MATCHES,
        description="Number of memories to retrieve from MemoryOp",
    )
    # Freeze to ensure the only mutation happens in either the agent state (which is
    # passed around) or in the internal Ops
    model_config = ConfigDict(frozen=True)

    @staticmethod
    def _format_memories(prompt: str, memories: Iterable[Memory]) -> str:
        return indent_xml(
            "\n".join([
                prompt.format(**m.model_dump(exclude={"run_id", "template"}))
                for m in memories
            ])
        )

    @staticmethod
    def _package_messages(
        msgs: list[Message], memory_prompt: str, use_memories: bool
    ) -> list[Message]:
        if use_memories:
            return [*msgs, Message(content=memory_prompt)]
        return msgs

    def __init__(self, memory_model: MemoryModel | None = None, **kwargs):
        super().__init__(**kwargs)
        self._query_factory_op = FxnOp[str](self.query_factory)
        self._memory_op = MemoryOp(memory_model)
        self._format_memory_op = FxnOp(self._format_memories)
        self._prompt_op = PromptOp(self.prompt)
        self._package_op = FxnOp(self._package_messages)

    @compute_graph()
    async def get_asv(
        self, agent_state: SimpleAgentState, obs: list[Message]
    ) -> tuple[OpResult[ToolRequestMessage], SimpleAgentState, float]:
        next_state = agent_state.get_next_state(obs)

        memories = await self._memory_op(
            query=await self._query_factory_op(next_state.messages),
            matches=self.num_memories,
        )
        packaged_messages = await self._package_op(
            next_state.messages,
            memory_prompt=await self._prompt_op(
                memories=await self._format_memory_op(self.memory_prompt, memories)
            ),
            use_memories=bool(memories.value),
        )
        result = cast(
            "OpResult[ToolRequestMessage]",
            await self._llm_call_op(
                await self._config_op(), msgs=packaged_messages, tools=next_state.tools
            ),
        )
        next_state.messages = [*next_state.messages, result.value]
        return result, next_state, 0.0



import pytest
from lmi import EmbeddingModel
from pytest_subtests import SubTests

from ldp.graph import Memory
from ldp.graph.memory import UIndexMemoryModel


@pytest.fixture(name="sample_memory")
def fixture_sample_memory() -> Memory:
    return Memory(
        query="sample string representation", output="observation", value=42.0
    )


class TestUIndexMemoryModel:
    def test_initialization_serialization(self, subtests: SubTests) -> None:
        with subtests.test(msg="default-model-specified"):
            model = UIndexMemoryModel()
            assert isinstance(model.embedding_model, EmbeddingModel), (
                "Default embedding model should be set"
            )
            model.model_dump()  # Check we can serialize

        with subtests.test(msg="nondefault-model-specified"):
            model_custom = UIndexMemoryModel(
                embedding_model=EmbeddingModel.from_name("text-embedding-3-small")
            )
            model_custom.model_dump()  # Check we can serialize

    @pytest.mark.asyncio
    async def test_add_then_get_memory(self, sample_memory: Memory) -> None:
        memory_model = UIndexMemoryModel(
            embedding_model=EmbeddingModel.from_name("text-embedding-3-small")
        )
        async with memory_model.safe_access_index() as index:
            assert len(index) == 0, "Should have no memories"
        await memory_model.add_memory(sample_memory)
        async with memory_model.safe_access_index() as index:
            assert len(index) == 1, "Should have one memory"
        assert memory_model.memories[0] == sample_memory
        result = await memory_model.get_memory("sample query", matches=1)
        assert len(result) == 1
        assert result[0] == sample_memory



from dataclasses import dataclass
from typing import Any

from aviary.core import Message, ToolRequestMessage, ToolResponseMessage
from aviary.message import EnvStateMessage

from ldp.utils import format_error_details, split_message_transitions


@dataclass
class MockResponse:
    status_code: int
    _json: dict | None = None
    _text: str = ""

    def json(self) -> dict[str, Any]:
        if self._json is None:
            raise ValueError("No JSON")
        return self._json

    @property
    def text(self) -> str:
        return self._text


class MockHTTPError(Exception):
    def __init__(self, status_code: int, detail: str | None = None, text: str = ""):
        self.response = MockResponse(
            status_code=status_code,
            _json={"detail": detail} if detail else None,
            _text=text,
        )
        super().__init__(f"HTTP {status_code}")


def test_format_basic_error():
    error = ValueError("something went wrong")
    details = format_error_details(error)
    assert details == "something went wrong"


def test_format_http_error_with_json():
    error = MockHTTPError(
        status_code=500,
        detail="Traceback:\n  File 'app.py', line 123\n  raise ValueError('oops')",
    )
    details = format_error_details(error)
    assert "Status code: 500" in details
    assert "Server Traceback:" in details
    assert "File 'app.py'" in details


def test_format_http_error_with_text():
    error = MockHTTPError(status_code=404, text="Not found")
    details = format_error_details(error)
    assert "Status code: 404" in details
    assert "Response body: Not found" in details


def test_split_message_transitions():
    """Test message breakdown with EnvStateMessage→ToolRequestMessage transitions."""
    messages = [
        Message(role="system", content="System prompt"),
        Message(role="user", content="Initial user message"),
        ToolRequestMessage(content="First tool request", tool_calls=[]),
        ToolResponseMessage(name="tool1", content="Tool response", tool_call_id="1"),
        EnvStateMessage(content="Environment state 1"),
        ToolRequestMessage(content="Second tool request", tool_calls=[]),
        ToolResponseMessage(name="tool2", content="Tool response 2", tool_call_id="2"),
        EnvStateMessage(content="Environment state 2"),
        ToolRequestMessage(content="Third tool request", tool_calls=[]),
        ToolResponseMessage(name="tool3", content="Tool response 3", tool_call_id="3"),
    ]

    result = split_message_transitions(messages)

    assert len(result) == 4

    # Block 1:
    assert len(result[0]) == 2
    assert result[0][0].role == "system"
    assert result[0][1].role == "user"

    # Block 2:
    assert len(result[1]) == 3
    assert isinstance(result[1][0], ToolRequestMessage)
    assert isinstance(result[1][1], ToolResponseMessage)
    assert isinstance(result[1][2], EnvStateMessage)

    # Block 3:
    assert len(result[2]) == 3
    assert isinstance(result[2][0], ToolRequestMessage)
    assert isinstance(result[2][1], ToolResponseMessage)
    assert isinstance(result[2][2], EnvStateMessage)

    # Block 4:
    assert len(result[3]) == 2
    assert isinstance(result[3][0], ToolRequestMessage)
    assert isinstance(result[3][1], ToolResponseMessage)



from collections.abc import Callable, Iterable
from typing import Any, Generic, TypeVar

from aviary.core import Message

from ldp.graph import ConfigOp, FxnOp, LLMCallOp, OpResult, compute_graph

TParsedMessage = TypeVar("TParsedMessage", bound=Message)


class ParsedLLMCallModule(Generic[TParsedMessage]):
    """Module for a processing-based tool selection, with a learnable configuration."""

    def __init__(
        self, llm_model: dict[str, Any], parser: Callable[..., TParsedMessage]
    ):
        self.config_op = ConfigOp[dict](config=llm_model)
        self.llm_call_op = LLMCallOp()
        self.parse_msg_op = FxnOp(parser)

    @compute_graph()
    async def __call__(
        self, messages: Iterable[Message], *parse_args, **parse_kwargs
    ) -> tuple[OpResult[TParsedMessage], Message]:
        raw_result = await self.llm_call_op(await self.config_op(), msgs=messages)
        return (
            await self.parse_msg_op(raw_result, *parse_args, **parse_kwargs),
            raw_result.value,
        )



import logging
from typing import cast

import torch
import torch.distributed as dist
from aviary.core import Message, Tool, ToolRequestMessage
from litellm.utils import token_counter
from pydantic import Field, field_validator

from ldp.agent import Agent, SimpleAgentState
from ldp.graph import OpResult
from ldp.graph.op_utils import compute_graph
from ldp.llms import prepend_sys
from ldp.nn.graph.llm_call_op import LocalLLMCallOp
from ldp.nn.handlers.chunking import TensorChunker
from ldp.nn.handlers.transformer_handler import (
    ParallelModeConfig,
    logits_to_logprobs,
)
from ldp.nn.lm_config import LMConfig as _LMConfig

logger = logging.getLogger(__name__)


class AgentLMConfig(_LMConfig):
    """Adds some additional configuration options for running an LM in an Op."""

    # async batching parameters
    batch_size: int = 1
    max_wait_interval: float = 0.1

    # distribution
    parallel_mode: ParallelModeConfig | None = None

    # sampling parameters
    temperature: float = 1.0
    max_new_tokens: int = 50
    parallel_tool_calls: bool = False

    llm_call_kwargs: dict = Field(
        default_factory=dict,
        description=(
            "Additional kwargs to pass to LocalLLMCallOp.forward. "
            "Note that the validator defaults top_k=None and top_p=1.0, which "
            "are better defaults than HF's."
        ),
        validate_default=True,
    )
    max_token_count: int | None = Field(
        default=None,
        description=(
            "If set, raise an error if the total tokens in the message history "
            "and tool description exceed this value."
        ),
    )

    @field_validator("llm_call_kwargs")
    @classmethod
    def validate_llm_call_kwargs(cls, v: dict) -> dict:
        return {"top_k": None, "top_p": 1.0} | v


# TODO: consider merging with SimpleAgent
class SimpleLocalLLMAgent(Agent[SimpleAgentState]):
    """Simple agent that can pick and invoke tools with a local language model."""

    def __init__(
        self,
        llm_model: AgentLMConfig,
        sys_prompt: str | None = None,
        hide_old_env_states: bool = False,
    ):
        self.llm_model = llm_model
        self.sys_prompt = sys_prompt
        self.hide_old_env_states = hide_old_env_states

        # Initialize the local LLM operation with configuration specifics
        self._llm_call_op = LocalLLMCallOp(
            model_config=self.llm_model,
            batch_size=self.llm_model.batch_size,
            max_wait_interval=self.llm_model.max_wait_interval,
            parallel_mode_config=self.llm_model.parallel_mode,
        )

    async def init_state(self, tools: list[Tool]) -> SimpleAgentState:
        """Initialize the state with provided tools."""
        return SimpleAgentState(
            tools=tools, hide_old_env_states=self.hide_old_env_states
        )

    @compute_graph()
    async def get_asv(
        self, agent_state: SimpleAgentState, obs: list[Message]
    ) -> tuple[OpResult[ToolRequestMessage], SimpleAgentState, float]:
        """Generates an action-state-value tuple using the local LLM model."""
        next_state = agent_state.get_next_state(obs)

        # If system prompt exists, prepend it; otherwise, use the existing messages
        messages = (
            prepend_sys(next_state.messages, sys_content=self.sys_prompt)
            if self.sys_prompt is not None
            else next_state.messages
        )

        self._validate_token_count(messages, next_state.tools)

        # Execute the LLM operation call
        result = await self._llm_call_op(
            xi=messages,
            temperature=self.llm_model.temperature,
            max_new_tokens=self.llm_model.max_new_tokens,
            tools=next_state.tools,
            **self.llm_model.llm_call_kwargs,
        )

        # Type-checking for expected output
        if not isinstance(result.value, ToolRequestMessage):
            raise TypeError(
                f"Expected ToolRequestMessage, got {type(result.value)}:"
                f" {result.value}, history: {messages}"
            )

        # Update state messages with result and return the new state
        next_state.messages = [*next_state.messages, result.value]
        return cast("OpResult[ToolRequestMessage]", result), next_state, 0.0

    def _validate_token_count(self, messages: list[Message], tools: list[Tool]):
        """Asserts token count for the trajectory is within the limit."""
        if self.llm_model.max_token_count is None:
            return
        messages_for_tokenizer = self._llm_call_op.prep_messages_for_tokenizer(messages)
        tools_for_tokenizer = self._llm_call_op.prep_tools_for_tokenizer(tools)

        total_tokens = token_counter(
            model=self.llm_model.model,
            messages=messages_for_tokenizer,
            tools=tools_for_tokenizer,  # type: ignore[arg-type]
        )
        if total_tokens > self.llm_model.max_token_count:
            logger.error(
                f"Token limit exceeded: {total_tokens} > {self.llm_model.max_token_count}"
            )
            raise ValueError(
                f"Token limit exceeded: {total_tokens} > {self.llm_model.max_token_count}"
            )

    # TODO: maybe remove these recomputation methods. I added them to debug some things. But idk,
    # maybe they'll come in handy later.
    @staticmethod
    def recompute_logprobs_forward(
        handler, input_ids: torch.Tensor, attention_mask: torch.Tensor
    ) -> torch.Tensor:
        num_chunks = dist.get_world_size()
        curr_chunk = dist.get_rank()
        chunker = TensorChunker(num_chunks=num_chunks)
        split_args_list, *_ = chunker.chunkify(input_ids, attention_mask)
        input_ids, attention_mask = split_args_list[curr_chunk]

        input_ids, attention_mask = (
            input_ids.to(handler.module.device),
            attention_mask.to(handler.module.device),
        )
        with torch.no_grad():
            handler.module.eval()
            out = handler.module(input_ids=input_ids, attention_mask=attention_mask)
        return out.logits.float()

    def recompute_logprobs(self, messages_: list[Message], tools_: list[Tool]):
        op = self._llm_call_op
        handler = op.model_handler
        messages = op.prep_messages_for_tokenizer(messages_)
        tools = op.prep_tools_for_tokenizer(tools_)

        encoded = handler.tokenizer.apply_chat_template(  # type: ignore[attr-defined]
            [messages],
            tools=tools,
            return_tensors="pt",
            return_dict=True,
            return_assistant_tokens_mask=True,
            tokenizer_kwargs={
                "return_special_tokens_mask": True,
            },
        )
        am = torch.tensor(encoded["assistant_masks"])
        action_start = (1 - am[0]).argwhere()[-1].item() + 1

        out = handler.wrap_func(
            SimpleLocalLLMAgent.recompute_logprobs_forward,
            worker_agg_fn=torch.concatenate,
        )(encoded["input_ids"], encoded["attention_mask"]).squeeze(0)

        logprobs_reco = logits_to_logprobs(
            out[:-1], encoded["input_ids"].squeeze(0)[1:]
        )
        # +3 to remove the assistant prompt tokens
        return logprobs_reco[action_start + 3 :]



from .common_ops import (
    ConfigOp,
    EmbeddingOp,
    FxnOp,
    IdentityOp,
    LLMCallOp,
    MemoryOp,
    PromptOp,
)
from .loss_ops import MSELossOp
from .memory import Memory, MemoryModel
from .op_utils import (
    CallID,
    compute_graph,
    eval_mode,
    get_call_id,
    get_run_id,
    get_training_mode,
    op_call,
    set_training_mode,
    train_mode,
)
from .ops import Op, OpCtx, OpResult

__all__ = [
    "CallID",
    "ConfigOp",
    "EmbeddingOp",
    "FxnOp",
    "IdentityOp",
    "LLMCallOp",
    "MSELossOp",
    "Memory",
    "MemoryModel",
    "MemoryOp",
    "Op",
    "OpCtx",
    "OpResult",
    "PromptOp",
    "compute_graph",
    "eval_mode",
    "get_call_id",
    "get_run_id",
    "get_training_mode",
    "op_call",
    "set_training_mode",
    "train_mode",
]



import asyncio
import itertools
import logging
import time
import traceback
import uuid
from collections import Counter
from collections.abc import Callable, Iterator, Sequence
from contextlib import contextmanager, nullcontext
from typing import Any, TypeVar, overload

from aviary.core import Environment, Message
from tqdm.asyncio import tqdm

from ldp.agent import Agent
from ldp.data_structures import Trajectory, Transition
from ldp.utils import format_error_details

from .callbacks import Callback

logger = logging.getLogger(__name__)


TEnv = TypeVar("TEnv", bound=Environment)


class CaughtError(Exception):
    """Base class for reraised exceptions when catching is enabled."""

    def __init__(self, original_exc: Exception):
        super().__init__(str(original_exc))
        self.original_exc = original_exc
        self.original_traceback = (
            original_exc.__traceback__
        )  # Store the original traceback

    exc_type = "undefined"

    def __str__(self):
        # Format the original exception with its traceback
        original_trace = "".join(
            traceback.format_exception(
                type(self.original_exc), self.original_exc, self.original_traceback
            )
        )
        return f"{self.exc_type} error: {super().__str__()}\nOriginal traceback:\n{original_trace}"


class AgentError(CaughtError):
    exc_type = "agent"


class EnvError(CaughtError):
    exc_type = "env"


@contextmanager
def reraise_exc_as(reraise: type[CaughtError], enabled: bool) -> Iterator[None]:
    """Context manager that reraises exceptions as a custom CaughtError type if enabled."""
    try:
        yield
    except Exception as e:
        if enabled:
            logger.info(f"Reraising {reraise.exc_type} exception.")
            raise reraise(e) from None
        raise


class Timer:
    """Tracks time spent in named operations."""

    def __init__(self):
        self.info: dict[str, float] = {}

    @contextmanager
    def __call__(self, name: str):
        start_time = time.monotonic()
        try:
            yield
        finally:
            self.info[f"time_elapsed_{name}"] = time.monotonic() - start_time


class RolloutManager:
    def __init__(
        self,
        agent: Agent,
        catch_agent_failures: bool = True,
        catch_env_failures: bool = True,
        callbacks: Sequence[Callback] | None = None,
        concurrency_limit: int | None = None,
    ):
        self.agent = agent

        self.catch_agent_failures = catch_agent_failures
        self.catch_env_failures = catch_env_failures

        self.concurrency_limiter = (
            asyncio.Semaphore(concurrency_limit) if concurrency_limit else nullcontext()
        )

        self.traj_buffer: dict[str, Trajectory] = {}
        self.callbacks = callbacks or []

    @overload
    async def sample_trajectories(  # noqa: D418
        self,
        environment_factory: Callable[[], TEnv],
        batch_size: int = 1,
        max_steps: int | None = None,
    ) -> list[tuple[Trajectory, TEnv]]:
        """Run rollouts in parallel, using a factory to construct environments.

        We will construct `batch_size` environments and run rollouts on each of them.
        If `max_steps` is set, rollouts will be truncated at this value. If a rollout
        has fewer than `max_steps`, then a new environment will be constructed and another
        rollout will be started until `max_steps` is reached.

        Args:
            environment_factory: A no-argument callable that returns
                an environment instance
            batch_size (int, optional): Defaults to 1.
            max_steps (int | None, optional): Max steps per rollout. Defaults to None (see above).

        Returns:
            list[tuple[Trajectory, Environment]]: A list of (trajectory, environment) tuples: one per rollout.
        """

    @overload
    async def sample_trajectories(  # noqa: D418
        self,
        environments: Sequence[Environment],
        max_steps: int | None = None,
    ) -> list[Trajectory]:
        """Run rollouts in parallel on a list of provided environments.

        Args:
            environments: A list of environments to run rollouts on.
            max_steps: Max steps per rollout. Defaults to None, in which case the rollouts are run
                until environment returns done.
            summarize_exceptions: Whether to collect exceptions and show a summary at the end.
                Defaults to True. If False, exceptions will be logged immediately as they occur
                during rollout.
        """

    async def sample_trajectories(self, **kwargs):
        if "environment_factory" in kwargs:
            assert "environments" not in kwargs, (
                "Cannot use environment_factory with environments"
            )

            return await self._sample_trajectories_from_env_factory(
                kwargs["environment_factory"],
                kwargs.get("batch_size", 1),
                kwargs.get("max_steps"),
                summarize_exceptions=kwargs.get("summarize_exceptions", False),
            )

        if "environments" in kwargs:
            assert "environment_factory" not in kwargs, (
                "Cannot use environments with environment_factory"
            )
            return await self._sample_trajectories_from_envs(
                kwargs["environments"],
                kwargs.get("max_steps"),
                summarize_exceptions=kwargs.get(
                    "summarize_exceptions",
                    False,
                ),
            )

        raise TypeError(
            "sample_trajectories() missing required "
            "arguments 'environment_factory' or 'environments'"
        )

    async def _sample_trajectories_from_env_factory(
        self,
        environment_factory: Callable[[], Environment],
        batch_size: int = 1,
        max_steps: int | None = None,
        *,
        summarize_exceptions: bool = False,
    ) -> list[tuple[Trajectory, Environment]]:
        self.traj_buffer.clear()
        exception_counter: Counter = Counter()

        async def rollout_with_args(idx: int, **rollout_kwargs):
            return idx, await self._rollout(**rollout_kwargs), rollout_kwargs

        accumulated_steps = [0] * batch_size
        total_trajectories = 0  # Counter for completed trajectories

        # submit initial batch of tasks
        tasks = [
            asyncio.create_task(
                rollout_with_args(
                    idx,
                    traj_id=uuid.uuid4().hex,
                    env=environment_factory(),
                    max_steps=max_steps,
                    summarize_exceptions=summarize_exceptions,
                )
            )
            for idx in range(batch_size)
        ]

        results = []
        with tqdm(
            desc="Rollouts",
            unit="rollout",
            ncols=0,
            disable=not summarize_exceptions,
        ) as pbar:
            while tasks:
                done, pending = await asyncio.wait(
                    tasks, return_when=asyncio.FIRST_COMPLETED
                )
                new_tasks = []
                for task in done:
                    idx, traj, kwargs = await task
                    results.append((traj, kwargs["env"]))
                    total_trajectories += 1
                    pbar.update(1)

                    steps_in_traj = len(traj.steps)
                    accumulated_steps[idx] += steps_in_traj

                    # Check for exceptions in this trajectory
                    if traj.steps and traj.steps[-1].metadata.get("exception"):
                        exc_str: str = str(traj.steps[-1].metadata["exception"])[
                            :500
                        ].replace('"', "'")
                        exception_counter[exc_str] += 1
                        num_exceptions = sum(exception_counter.values())
                        pbar.set_postfix({"num_exceptions": num_exceptions})

                    if (
                        max_steps is not None
                        and (remaining_steps := max_steps - accumulated_steps[idx]) > 0
                    ):
                        # submit another task if we haven't reached max_steps
                        new_task = asyncio.create_task(
                            rollout_with_args(
                                idx,
                                traj_id=uuid.uuid4().hex,
                                env=environment_factory(),
                                max_steps=remaining_steps,
                                summarize_exceptions=summarize_exceptions,
                            )
                        )
                        new_tasks.append(new_task)

                tasks = list(pending) + new_tasks

        # Final summary of exceptions (if any)
        if exception_counter and summarize_exceptions:
            summary = ["Caught exceptions:", "Count  Exception"]
            summary.extend(
                f"{count:<6d} {exc:<50s}" for exc, count in exception_counter.items()
            )
            logger.info("\n".join(summary))

        return results

    async def _sample_trajectories_from_envs(
        self,
        environments: Sequence[Environment],
        max_steps: int | None = None,
        *,
        summarize_exceptions: bool = False,
    ) -> list[Trajectory]:
        self.traj_buffer.clear()
        exception_counter: Counter = Counter()

        traj_ids = [uuid.uuid4().hex for _ in environments]

        # Create all tasks first
        tasks = [
            asyncio.create_task(
                self._rollout(
                    traj_id,
                    env,
                    max_steps=max_steps,
                    summarize_exceptions=summarize_exceptions,
                )
            )
            for traj_id, env in zip(traj_ids, environments, strict=True)
        ]

        with tqdm(
            total=len(tasks),
            desc="Rollouts",
            unit="rollout",
            ncols=0,
            disable=not summarize_exceptions,
        ) as pbar:
            for task in asyncio.as_completed(tasks):
                trajectory = await task
                pbar.update(1)
                # Check if this trajectory ended with an exception
                if trajectory.steps:
                    last_step = trajectory.steps[-1]
                    if last_step.metadata.get("exception"):
                        # We'll keep it short but still have something to categorize
                        exc_str: str = str(last_step.metadata["exception"])[
                            :500
                        ].replace('"', "'")
                        exception_counter[exc_str] += 1
                        num_exceptions = sum(exception_counter.values())
                        pbar.set_postfix({"num_exceptions": num_exceptions})

        # Final summary of exceptions (if any)
        if exception_counter and summarize_exceptions:
            summary = ["Caught exceptions:", "Count  Exception"]
            summary.extend(
                f"{count:<6d} {exc:<50s}" for exc, count in exception_counter.items()
            )
            logger.info("\n".join(summary))

        return [self.traj_buffer[traj_id] for traj_id in traj_ids]

    async def _rollout(
        self,
        traj_id: str,
        env: Environment,
        max_steps: int | None,
        *,
        summarize_exceptions: bool = False,
    ) -> Trajectory:
        trajectory = await Trajectory.from_env(env, traj_id=traj_id)
        timer = Timer()

        async def store_step(step: Transition):
            await asyncio.gather(*[
                callback.after_transition(traj_id, self.agent, env, step)
                for callback in self.callbacks
            ])
            trajectory.steps.append(step)

        # Set default values to store in the buffer in case reset/init_state fail
        obs: list[Message] = []
        agent_state: Any = None

        try:
            await asyncio.gather(*[
                c.before_rollout(traj_id, env) for c in self.callbacks
            ])

            with reraise_exc_as(EnvError, enabled=self.catch_env_failures):
                obs, tools = await env.reset()
            await asyncio.gather(*[
                c.after_env_reset(traj_id, obs, tools) for c in self.callbacks
            ])

            with reraise_exc_as(AgentError, enabled=self.catch_agent_failures):
                agent_state = await self.agent.init_state(tools)
            await asyncio.gather(*[
                c.after_agent_init_state(traj_id, agent_state) for c in self.callbacks
            ])

            for timestep in itertools.count():
                step = await self._take_step(
                    timestep, traj_id, env, agent_state, obs, timer
                )

                if timestep + 1 == max_steps and not step.done:
                    # Mark as truncated if we hit max_steps and the state is not terminal.
                    # Do it before store_step(), so that callbacks can access this info
                    step.truncated = True

                # We assume the below won't throw a CaughtError
                await store_step(step)

                # set things up for the next iteration
                agent_state = step.next_agent_state
                obs = step.next_observation

                if step.done or step.truncated:
                    break

        except CaughtError as e:
            # NOTE: This trajectory should not be used for regular training.
            # We save the last transition here for debugging, etc.
            if not summarize_exceptions:
                error_details = format_error_details(e.original_exc)
                logger.exception(f"Exception in rollout {traj_id}:\n{error_details}")

            await store_step(
                Transition(
                    timestep=len(trajectory.steps),
                    agent_state=agent_state,
                    next_agent_state=None,
                    observation=obs,
                    next_observation=[],
                    action=None,
                    done=True,
                    metadata={"exception": repr(e.original_exc)} | timer.info,
                )
            )
        finally:
            with reraise_exc_as(EnvError, enabled=self.catch_env_failures):
                await env.close()

        self.traj_buffer[traj_id] = trajectory
        return trajectory

    async def _take_step(
        self,
        timestep: int,
        traj_id: str,
        env: Environment,
        agent_state: Any,
        obs: list[Message],
        timer: Timer | None = None,
    ) -> Transition:
        timer = timer or Timer()

        async with self.concurrency_limiter:
            with timer("before_transition"):
                await asyncio.gather(*[
                    callback.before_transition(
                        traj_id, self.agent, env, agent_state, obs
                    )
                    for callback in self.callbacks
                ])

            with (
                timer("agent_get_asv"),
                reraise_exc_as(AgentError, enabled=self.catch_agent_failures),
            ):
                (
                    action,
                    next_agent_state,
                    value,
                ) = await self.agent.get_asv(agent_state, obs)

            with timer("after_agent_get_asv"):
                await asyncio.gather(*[
                    callback.after_agent_get_asv(
                        traj_id, action, next_agent_state, value
                    )
                    for callback in self.callbacks
                ])

            with (
                timer("env_step"),
                reraise_exc_as(EnvError, enabled=self.catch_env_failures),
            ):
                next_obs, reward, done, trunc = await env.step(action.value)
            with timer("after_env_step"):
                await asyncio.gather(*[
                    callback.after_env_step(traj_id, next_obs, reward, done, trunc)
                    for callback in self.callbacks
                ])

            return Transition(
                timestep=timestep,
                agent_state=agent_state,
                next_agent_state=next_agent_state,
                action=action,
                reward=reward,
                value=value,
                observation=obs,
                next_observation=next_obs,
                done=done,
                truncated=trunc,
                metadata=timer.info,
            )



from importlib.metadata import version
from typing import ClassVar

import litellm
import pytest
from aviary.core import (
    DummyEnv,
    DummyEnvState,
    Message,
    Tool,
    ToolCall,
    ToolRequestMessage,
    ToolResponseMessage,
)
from lmi import CommonLLMNames

from ldp.agent import SimpleAgent


class ParallelizedDummyEnv(DummyEnv):
    def __init__(self, right_hand_broken: bool = False):
        super().__init__()
        self.right_hand_broken = right_hand_broken

    RIGHT_HAND_BROKEN_MESSAGE: ClassVar[str] = "Right hand is broken."

    async def reset(self) -> tuple[list[Message], list[Tool]]:
        def move_right_hand(
            distance: int,  # noqa: ARG001
            state: DummyEnvState,
        ) -> None:
            """
            Move your right hand forward or backward.

            Args:
                distance: Integer distance to move (mm), where forward is positive.
                state: Current state.
            """
            if self.right_hand_broken:  # Use this to test tool errors
                raise RuntimeError(self.RIGHT_HAND_BROKEN_MESSAGE)
            state.reward += 1

        def move_left_hand(
            distance: int,  # noqa: ARG001
            state: DummyEnvState,
        ) -> None:
            """
            Move your left hand forward or backward.

            Args:
                distance: Integer distance to move (mm), where forward is positive.
                state: Current state.
            """
            state.reward += 1

        def smile_and_wave(state: DummyEnvState) -> None:
            """
            Smile and wave.

            Args:
                state: Current state.
            """
            state.reward = 10
            state.done = True

        self.tools = [
            Tool.from_function(move_left_hand),
            Tool.from_function(move_right_hand),
            Tool.from_function(smile_and_wave),
        ]
        self.state = type(self).State(
            messages=[
                Message(
                    role="user",
                    content=(
                        "You are the president of the United States of America."
                        " Please move both hands at the same time, and then smile"
                        " and wave."
                    ),
                )
            ]
        )
        return self.state.messages, self.tools


class TestParallelism:
    @pytest.mark.asyncio
    @pytest.mark.vcr
    async def test_SimpleAgent_can_parallel_call(self) -> None:
        env = ParallelizedDummyEnv()
        obs, tools = await env.reset()
        agent = SimpleAgent()

        # Check parallel tool calls
        action, agent_state, _ = await agent.get_asv(  # noqa: RUF059
            await agent.init_state(tools=tools), obs
        )
        selected_tools: set[str] = {tc.function.name for tc in action.value.tool_calls}
        assert {
            "move_left_hand",
            "move_right_hand",
        } <= selected_tools, (
            f"Agent should've chosen tools in parallel, but it chose {selected_tools}"
        )

    @pytest.mark.parametrize(
        "model_name", [CommonLLMNames.ANTHROPIC_TEST.value, "gpt-4-turbo"]
    )
    @pytest.mark.asyncio
    async def test_exec_tool_calls_handling(self, model_name: str) -> None:
        env = ParallelizedDummyEnv(right_hand_broken=True)
        obs, tools = await env.reset()
        right_hand_tool = tools[1]
        agent = SimpleAgent(
            llm_model=SimpleAgent.model_fields["llm_model"].default
            | {"name": model_name}
        )
        agent_state = await agent.init_state(tools=tools)

        # 1. Let's DIY create a ToolRequestMessage for test determinism
        request_msg = ToolRequestMessage(
            content="stub", tool_calls=[ToolCall.from_tool(right_hand_tool, distance=5)]
        )
        agent_state.messages.extend([*obs, request_msg])

        # 2. Okay, our hand was broken, let's handle it DIY-style
        try:
            obs, *_ = await env.step(action=request_msg)
        except RuntimeError as exc:
            obs = [
                Message(
                    content=f"Failed to execute tools with message:\n{exc}", role="tool"
                )
            ]
        else:
            raise AssertionError("Should have blown up per the test logic.")

        # 2. Well, it looks like both Anthropic and OpenAI don't like DIY-style
        #    (using a bare Message) because they expect a tool call ID and tool name
        with pytest.raises(
            litellm.BadRequestError,
            match=(
                "invalid" if version(litellm.__name__) < "1.45.0" else "tool_call_id"
            ),
        ):
            await agent.get_asv(agent_state, obs)

        # 3. Alright, let's check the agent doesn't blow up if we use a
        #    ToolResponseMessage as Anthropic and OpenAI expect
        await agent.get_asv(
            agent_state,
            ToolResponseMessage.from_request(request_msg, contents=["null"]),  # type: ignore[arg-type]
        )

        # 4. Now that we have confirmed that, let's make sure exec_tool_calls
        #    can automate this for us
        obs = await env.exec_tool_calls(  # type: ignore[assignment]
            message=request_msg, state=env.state, handle_tool_exc=True
        )
        (failure_tool_response,) = obs
        assert isinstance(failure_tool_response, ToolResponseMessage)
        assert env.RIGHT_HAND_BROKEN_MESSAGE in failure_tool_response.content
        await agent.get_asv(agent_state, obs)



from collections.abc import Sequence
from unittest.mock import patch

import litellm
import pytest
from aviary.core import DummyEnv, TaskDataset

from ldp.agent import MemoryAgent, SimpleAgent
from ldp.alg import (
    Callback,
    Evaluator,
    EvaluatorConfig,
    MeanMetricsCallback,
    OfflineTrainer,
    OfflineTrainerConfig,
    OnlineTrainer,
    OnlineTrainerConfig,
    StoreTrajectoriesCallback,
)
from ldp.alg.datasets import (  # noqa: F401  # Force TASK_DATASET_REGISTRY update
    DummyTaskDataset,
)
from ldp.alg.optimizer import default_optimizer_factory
from ldp.data_structures import Trajectory
from ldp.graph import OpCtx


@pytest.mark.vcr
@pytest.mark.asyncio
@pytest.mark.parametrize("clear_ctx_at_each_iter", [True, False])
async def test_online_trainer(clear_ctx_at_each_iter: bool) -> None:
    agent = MemoryAgent()
    opt = default_optimizer_factory(agent)
    dataset = TaskDataset.from_name("dummy")
    dummy_callback = DummyCallback()
    metrics_callback = MeanMetricsCallback(train_dataset=dataset, track_tool_usage=True)

    train_conf = OnlineTrainerConfig(
        batch_size=1,
        num_train_iterations=1,
        max_rollout_steps=1,
        num_eval_iterations=1,
        num_rollouts_per_env=2,
        eval_every=1,
        clear_ctx_at_each_iter=clear_ctx_at_each_iter,
    )
    trainer = OnlineTrainer(
        config=train_conf,
        agent=agent,
        optimizer=opt,
        train_dataset=dataset,
        eval_dataset=dataset,
        callbacks=[dummy_callback, metrics_callback],
    )
    await trainer.train()

    for k, v in dummy_callback.counters.items():
        if k == "num_train_trajectories":
            # because num_rollouts_per_env==2
            assert v == 2
        elif "eval" in k:
            # eval is run 3 times: before training, during training, after training
            assert v == 3
        else:
            # after_{train_step,update} should be called once each
            assert v == 1
    assert metrics_callback.train_means["failures"] < 1, "Training should work"
    assert "tool_print_story" in metrics_callback.train_means

    if clear_ctx_at_each_iter:
        all(not ctx_data.data for ctx_data in OpCtx._CTX_REGISTRY.values())
    else:
        any(ctx_data.data for ctx_data in OpCtx._CTX_REGISTRY.values())


@pytest.mark.asyncio
@pytest.mark.parametrize("clear_ctx_at_each_iter", [True, False])
async def test_evaluator(clear_ctx_at_each_iter) -> None:
    agent = SimpleAgent()
    dataset = TaskDataset.from_name("dummy")
    metrics_callback = MeanMetricsCallback(eval_dataset=dataset)
    count_callback = DummyCallback()

    eval_conf = EvaluatorConfig(
        num_eval_iterations=1,
        clear_ctx_at_each_iter=clear_ctx_at_each_iter,
    )
    evaluator = Evaluator(
        config=eval_conf,
        agent=agent,
        dataset=dataset,
        callbacks=[metrics_callback, count_callback],
    )
    with patch.object(DummyEnv, "close", autospec=True) as spy_close:
        await evaluator.evaluate()

    spy_close.assert_awaited_once()
    assert isinstance(metrics_callback.eval_means["reward"], float)
    assert "tool_print_story" not in metrics_callback.eval_means

    for k, v in count_callback.counters.items():
        assert v == (1 if "eval" in k else 0)

    if clear_ctx_at_each_iter:
        all(not ctx_data.data for ctx_data in OpCtx._CTX_REGISTRY.values())
    else:
        any(ctx_data.data for ctx_data in OpCtx._CTX_REGISTRY.values())


@pytest.mark.asyncio
async def test_can_measure_evaluation_failure_rate() -> None:
    dataset = TaskDataset.from_name("dummy")
    metrics_callback = MeanMetricsCallback(eval_dataset=dataset)

    evaluator = Evaluator(
        config=EvaluatorConfig(
            batch_size=1, num_eval_iterations=1, max_rollout_steps=2
        ),
        agent=SimpleAgent(),
        dataset=dataset,
        callbacks=[metrics_callback],
    )
    with (
        patch.object(
            type(evaluator.agent), "get_asv", side_effect=litellm.APIError
        ) as mock_get_asv,
        patch.object(DummyEnv, "close", autospec=True) as spy_close,
    ):
        await evaluator.evaluate()  # Confirm this does not crash
    mock_get_asv.assert_awaited_once()
    assert metrics_callback.eval_means["failures"] == 1.0
    spy_close.assert_awaited_once()


@pytest.mark.vcr
@pytest.mark.asyncio
@pytest.mark.parametrize("clear_ctx_at_each_iter", [True, False])
async def test_offline_trainer(clear_ctx_at_each_iter: bool) -> None:
    # This is kind of a system test of getting trajectories from the evaluator
    # and then training on them "offline"
    agent = MemoryAgent()
    opt = default_optimizer_factory(agent)
    dataset = TaskDataset.from_name("dummy")
    traj_callback = StoreTrajectoriesCallback()

    evaluator = Evaluator(
        config=EvaluatorConfig(
            num_eval_iterations=1,
            clear_ctx_at_each_iter=clear_ctx_at_each_iter,
        ),
        agent=agent,
        dataset=dataset,
        callbacks=[traj_callback],
    )
    await evaluator.run()
    assert len(traj_callback.eval_trajectories) == 1

    count_callback = DummyCallback()
    metrics_callback = MeanMetricsCallback(train_dataset=dataset)
    trainer = OfflineTrainer(
        config=OfflineTrainerConfig(
            batch_size=1,
            clear_ctx_at_each_iter=clear_ctx_at_each_iter,
            num_epochs=2,
        ),
        agent=agent,
        optimizer=opt,
        train_trajectories=traj_callback.eval_trajectories,
        callbacks=[count_callback, metrics_callback],
    )
    await trainer.train()

    assert count_callback.counters == {
        "after_train_step": 2,
        "after_eval_step": 0,
        "after_eval_loop": 0,
        "after_update": 2,
        "num_train_trajectories": 2,
    }
    assert metrics_callback.train_means["failures"] < 1, "Training should work"

    if clear_ctx_at_each_iter:
        all(not ctx_data.data for ctx_data in OpCtx._CTX_REGISTRY.values())
    else:
        any(ctx_data.data for ctx_data in OpCtx._CTX_REGISTRY.values())


class DummyCallback(Callback):
    def __init__(self):
        self.counters = {
            "after_train_step": 0,
            "after_eval_step": 0,
            "after_eval_loop": 0,
            "after_update": 0,
            "num_train_trajectories": 0,
        }

    async def after_train_step(self, trajectories: Sequence[Trajectory]) -> None:
        self.counters["after_train_step"] += 1
        self.counters["num_train_trajectories"] += len(trajectories)

    async def after_eval_step(self, trajectories: Sequence[Trajectory]) -> None:
        self.counters["after_eval_step"] += 1

    async def after_eval_loop(self) -> None:
        self.counters["after_eval_loop"] += 1

    async def after_update(self) -> None:
        self.counters["after_update"] += 1



import asyncio

import pytest
from aviary.core import Message

import ldp.nn
from ldp.graph import OpResult


class TestLocalLLMCallOp:
    @pytest.mark.asyncio
    async def test_batching_consistent_results(self):
        """Tests that padding in batched calls does not affect the results (attention_mask is set correctly)."""
        model_config = ldp.nn.LMConfig(
            model="gpt2", device="cpu", dtype=ldp.nn.TorchDType.fp32
        )
        local_llm_call_op = ldp.nn.LocalLLMCallOp(
            model_config, batch_size=2, max_wait_interval=1.0
        )

        messages = [
            Message(content=text)
            for text in ("Hello, how are you?", "Hello, how are you?")
        ]

        async def forward_batch() -> list[OpResult[Message]]:
            return await asyncio.gather(*[
                local_llm_call_op([msg], temperature=1.0, max_new_tokens=10)
                for msg in messages
            ])

        # First forward batch with seed set to 0
        ldp.nn.set_seed(0)
        results_first_call = await forward_batch()

        # Re-seed and forward batch again with padding to check for consistency
        ldp.nn.set_seed(0)
        messages[
            -1
        ].content = (
            "Some very long text that would create lots of padding in the batch."
        )
        results_second_call = await forward_batch()

        assert len(results_first_call) == len(results_second_call), (
            "Expected the number of results to match between the two calls."
        )
        assert results_first_call[0].value == results_second_call[0].value, (
            "Expected the results to match between the two calls, but got differing"
            " results."
        )



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from functools import wraps
from pathlib import Path
from typing import Any, Concatenate, ParamSpec, TypeVar

import torch
from torch import nn

from ldp.graph.async_torch import AsyncTorchModule

TReturn = TypeVar("TReturn")
TParams = ParamSpec("TParams")


class ModuleExecutionInterface(ABC):
    """Defines a common interface for interacting with PyTorch modules.

    Subclasses should implement @wrap_func and @wrap_afunc such that calling code can do:

    ```python
    exec_interface = MyExecutionInterface(...)


    @exec_interface.wrap_func
    def my_func(exec_interface, arg1, arg2):
        return exec_interface.module(arg1, arg2)


    @exec_interface.wrap_afunc
    async def my_afunc(exec_interface, arg1, arg2):
        return await exec_interface.module(arg1, arg2)
    ```
    """

    def wrap_afunc(
        self,
        # NOTE: type-hinting the first argument as Any so that subclasses can specialize
        # without typing errors.
        func: Callable[Concatenate[Any, TParams], Awaitable[TReturn]],
        **kwargs,
    ) -> Callable[TParams, Awaitable[TReturn]]:
        """Wraps a coroutine whose first argument is a ModuleExecutionInterface.

        By default, the coroutine is returned with the first argument
        set to this object, equivalent to `partial(func, self)`.

        Args:
            func: The coroutine to wrap.
            kwargs: Additional arguments - discarded, but may be used by subclasses.
        """

        @wraps(func)
        async def wrapped_func(*args, **kwargs):
            return await func(self, *args, **kwargs)

        return wrapped_func

    def wrap_func(
        self, func: Callable[Concatenate[Any, TParams], TReturn], **kwargs
    ) -> Callable[TParams, TReturn]:
        """Wraps a function whose first argument is a ModuleExecutionInterface.

        By default, the function is returned with the first argument
        set to this object, equivalent to `partial(func, self)`.

        Args:
            func: The function to wrap.
            kwargs: Additional arguments - discarded, but may be used by subclasses.
        """

        @wraps(func)
        def wrapped_func(*args, **kwargs):
            return func(self, *args, **kwargs)

        return wrapped_func

    @abstractmethod
    def load_state_dict(self, state_dict: dict[str, torch.Tensor]) -> None:
        """Load a provided state dict."""

    @abstractmethod
    def state_dict(self) -> dict[str, torch.Tensor]:
        """Return the current state dict."""

    @abstractmethod
    def load_checkpoint(self, ckpt: os.PathLike | str) -> None:
        """Load a checkpoint from the provided path."""

    @abstractmethod
    def save_checkpoint(self, ckpt: os.PathLike | str) -> None:
        """Save a checkpoint to the provided path."""

    def teardown(self):
        """Teardown any resources used by the interface."""


class ModuleHandler(ModuleExecutionInterface):
    """Base class for an execution interface that holds a PyTorch model.

    Subclasses can wrap calls to the model for specific use-cases, such as
    batching async calls, remote execution, etc.
    """

    def __init__(self, module: nn.Module):
        self.module: nn.Module = module

        # A container for objects that are local to the current process.
        # Can be used for something like an optimizer, if self.module is
        # sharded. The handler has to own these objects because they may
        # be created remotely (e.g. w/ Dask) and cannot be pickled.
        self.process_local_objects: dict[str, Any] = {}

    def state_dict(self) -> dict[str, torch.Tensor]:
        return self.module.state_dict()

    def load_state_dict(self, state_dict: dict[str, torch.Tensor]) -> None:
        self.module.load_state_dict(state_dict)

    def save_checkpoint(self, ckpt: os.PathLike | str, **kwargs) -> None:
        ckpt = Path(ckpt)
        if ckpt.is_dir():
            ckpt /= "model.pt"
        torch.save(self.state_dict(), ckpt)

    def load_checkpoint(self, ckpt: os.PathLike | str) -> None:
        ckpt = Path(ckpt)
        if ckpt.is_dir():
            ckpt /= "model.pt"
        self.load_state_dict(torch.load(ckpt))


class AsyncModuleHandler(ModuleHandler):
    def __init__(self, module: nn.Module, **kwargs):
        super().__init__(module)
        self.async_module = AsyncTorchModule(module, **kwargs)

    async def __call__(self, **kwargs):
        return await self.async_module(**kwargs)



from lmi import LiteLLMModel as LLMModel
from lmi import (
    LLMResult,
    sum_logprobs,
    validate_json_completion,
)
from lmi.embeddings import (
    EmbeddingModel,
    EmbeddingModes,
    HybridEmbeddingModel,
    LiteLLMEmbeddingModel,
    SparseEmbeddingModel,
)
from lmi.exceptions import (
    JSONSchemaValidationError,
)

from .prompts import (
    append_to_messages,
    append_to_sys,
    indent_xml,
    prepend_sys,
    prepend_sys_and_append_sys,
)

__all__ = [
    "EmbeddingModel",
    "EmbeddingModes",
    "HybridEmbeddingModel",
    "JSONSchemaValidationError",
    "LLMModel",
    "LLMResult",
    "LiteLLMEmbeddingModel",
    "SparseEmbeddingModel",
    "append_to_messages",
    "append_to_sys",
    "indent_xml",
    "prepend_sys",
    "prepend_sys_and_append_sys",
    "sum_logprobs",
    "validate_json_completion",
]



from enum import StrEnum
from logging import getLogger
from typing import Any, Self, TypeVar, no_type_check

import torch
from pydantic import BaseModel, ConfigDict, Field, model_validator
from transformers import (
    AutoModelForCausalLM,
    AutoModelForSequenceClassification,
    AutoTokenizer,
    BitsAndBytesConfig,
    PreTrainedModel,
    PreTrainedTokenizer,
)

logger = getLogger(__name__)
TModel = TypeVar("TModel", bound=PreTrainedModel)


class TorchDType(StrEnum):
    bf16 = "bfloat16"
    fp16 = "float16"
    fp32 = "float32"
    auto = "auto"


class LMConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    model: str = Field(
        description=(
            "Name of the model to load. Must be available "
            "on the Huggingface Hub or the path to a local directory."
        )
    )
    load_args: dict[str, Any] = Field(
        default_factory=dict,
        description="Passed as Model.from_pretrained(self.model, **load_args)",
    )
    tokenizer_args: dict[str, Any] = Field(
        default_factory=dict,
        description=(
            "Passed as AutoTokenizer.from_pretrained(self.model, **tokenizer_args)"
        ),
    )
    chat_template: str | None = Field(
        default=None,
        description=(
            "Name of a jinja file defining a chat template. "
            "Leave as None to not use a chat template."
        ),
    )

    device: str | int | None = None
    dtype: TorchDType = Field(
        default=TorchDType.auto,
        description=(
            "Will pass torch_dtype=getattr(torch, self.dtype) if dtype is not auto."
        ),
    )
    pad_token: str | None = Field(
        default=None,
        description=(
            "If set, will override the pad token in the tokenizer. "
            "Must be a valid token already in the vocabulary. Should be primarily "
            "used for tokenizers that do not predefine a pad token; will throw a "
            "warning otherwise."
        ),
    )
    gradient_checkpointing: bool = False
    compile: bool = False

    # private attribute
    _loaded_model_name: str | None = None

    @model_validator(mode="after")
    def check_load_args(self) -> Self:
        if (
            self.device != "cpu"
            and torch.cuda.is_available()
            and self.dtype in {TorchDType.bf16, TorchDType.fp16}
        ):
            # FA2 is a good default if provided
            self.load_args.setdefault("attn_implementation", "flash_attention_2")
        if "torch_dtype" in self.load_args:
            raise ValueError("Do not set torch_dtype in load_args. Use dtype instead.")
        return self

    def get_causal_lm(self) -> tuple[PreTrainedTokenizer, PreTrainedModel]:
        return self.get_model(AutoModelForCausalLM)

    def get_regression_lm(
        self,
    ) -> tuple[PreTrainedTokenizer, PreTrainedModel]:
        # num_labels=1 puts it in regression mode (MSE loss, single output)
        # https://huggingface.co/docs/transformers/v4.44.2/en/model_doc/auto#transformers.AutoModelForSequenceClassification
        return self.get_classification_lm(num_labels=1)

    def get_classification_lm(
        self, num_labels: int
    ) -> tuple[PreTrainedTokenizer, PreTrainedModel]:
        return self.get_model(AutoModelForSequenceClassification, num_labels=num_labels)

    # huggingface autotypes make type annotations messy, so disable mypy
    # for this function
    @no_type_check
    def get_model(
        self,
        model_cls: type[TModel],
        _compile_enabled: bool = True,
        **kwargs,
    ) -> tuple[PreTrainedTokenizer, TModel]:
        assert self._loaded_model_name is not None, (
            "Call resolve_model_location() before get_*() methods."
        )
        model = self._load_pretrained_model(
            self._loaded_model_name, model_cls, **kwargs
        )
        tokenizer = self._load_tokenizer(self._loaded_model_name)

        # Make consistent in case _load_tokenizer changed the pad token
        model.config.pad_token_id = tokenizer.pad_token_id

        if self.gradient_checkpointing:
            model.gradient_checkpointing_enable()  # will raise if not supported

        if _compile_enabled and self.compile:
            model = torch.compile(model)

        logger.debug(f"Model:\n{model}")

        return tokenizer, model

    def get_tokenizer(self) -> PreTrainedTokenizer:
        return self._load_tokenizer(self._loaded_model_name or self.model)

    @no_type_check
    def _load_tokenizer(self, model_name: str) -> PreTrainedTokenizer:
        logger.info(f"Loading tokenizer from {model_name}")
        tokenizer = AutoTokenizer.from_pretrained(model_name, **self.tokenizer_args)
        tokenizer.padding_side = "right"

        if self.pad_token is not None:
            if self.pad_token not in tokenizer.vocab:
                raise ValueError(
                    f"pad_token {self.pad_token!r} not in tokenizer vocabulary"
                )
            if tokenizer.pad_token is not None:
                logger.warning(
                    f"Overriding tokenizer pad token {tokenizer.pad_token!r} with"
                    f" {self.pad_token!r}"
                )
            tokenizer.pad_token = self.pad_token

        return tokenizer

    @no_type_check
    def _load_pretrained_model(
        self, model_name: str, model_cls: type[TModel], **kwargs
    ) -> TModel:
        kwargs.update(self.load_args)
        if "quantization_config" in kwargs:
            kwargs["quantization_config"] = BitsAndBytesConfig(
                **kwargs["quantization_config"]
            )
        if self.dtype != TorchDType.auto:
            kwargs["torch_dtype"] = getattr(torch, self.dtype.value)

        if self.device is not None:
            device_map = self.device
            logger.info(f"Loading model from {model_name} to {self.device}")
        else:
            device_map = None
            logger.info(f"Loading model from {model_name}")
        model = model_cls.from_pretrained(
            model_name,
            device_map=device_map,
            **kwargs,
        )

        if "torch_dtype" in kwargs:
            # ValueHead models put the value head in fp32, so homogenize here
            model = model.to(kwargs["torch_dtype"])

        return model

    def resolve_model_location(self, is_main_process: bool = True):
        """This method can be overridden in subclasses to load models from custom storage.

        This class automatically supports models in (a) Huggingface Hub or (b) the local
        filesystem.
        """
        self._loaded_model_name = self.model



import ast
import re
import textwrap
from collections.abc import Iterable
from enum import StrEnum
from typing import Any, cast

from aviary.core import (
    MalformedMessageError,
    Message,
    Messages,
    Tool,
    ToolCall,
    ToolRequestMessage,
)
from aviary.message import EMPTY_CONTENT_BASE_MSG

from ldp.graph import FxnOp, LLMCallOp, OpResult, PromptOp, compute_graph
from ldp.llms import prepend_sys

from .llm_call import ParsedLLMCallModule

# These prompts are meant to be used with ReActModuleSinglePrompt
_DEFAULT_SINGLE_PROMPT_TEMPLATE = textwrap.dedent(
    """    Answer the following questions as best you can. You have access to the following tools:

    {{tools}}

    Use the following format:

    {fields}
    ... (this {fields_description} can repeat N times)

    Example:

    {example}"""
)
REACT_DEFAULT_SINGLE_PROMPT_TEMPLATE = _DEFAULT_SINGLE_PROMPT_TEMPLATE.format(
    fields=(
        "Thought: you should always think about what to do"
        "\nAction: the action to take, should be one of [{tool_names}]"
        "\nAction Input: comma separated list of inputs to action as python tuple"
        "\nObservation: the result of the action"
    ),
    fields_description="Thought/Action/Action Input/Observation",
    example=(
        "Thought: I need to use the get_weather tool"
        "\nAction: get_weather"
        '\nAction Input: "New York", 7'
        "\nObservation: The 7 day forecast for New York is [...]"
    ),
)
ACT_DEFAULT_SINGLE_PROMPT_TEMPLATE = _DEFAULT_SINGLE_PROMPT_TEMPLATE.format(
    fields=(
        "Action: the action to take, should be one of [{tool_names}]"
        "\nAction Input: comma separated list of inputs to action as python tuple"
        "\nObservation: the result of the action"
    ),
    fields_description="Action/Action Input/Observation",
    example=(
        "Action: get_weather"
        '\nAction Input: "New York", 7'
        "\nObservation: The 7 day forecast for New York is [...]"
    ),
)


# And these with ReActModule
_DEFAULT_PROMPT_TEMPLATE = textwrap.dedent(
    """    Answer the following questions as best you can, using the provided tools.

    Use the following format:

    {fields}
    ... (this {fields_description} can repeat N times)

    Example:

    {example}"""
)
REACT_DEFAULT_PROMPT_TEMPLATE = _DEFAULT_PROMPT_TEMPLATE.format(
    fields=(
        "Thought: you should always think about what to do"
        "\nAction: the action to take,"
        " should be one of the provided tools with necessary arguments"
        "\nObservation: the result of the action"
    ),
    fields_description="Thought/Action/Observation",
    example=(
        "Thought: I need to use the get_weather tool"
        '\nAction: get_weather("New York", 7)'
        "\nObservation: The 7 day forecast for New York is [...]"
    ),
)
ACT_DEFAULT_PROMPT_TEMPLATE = _DEFAULT_PROMPT_TEMPLATE.format(
    fields=(
        "Action: the action to take,"
        " should be one of the provided tools with necessary arguments"
        "\nObservation: the result of the action"
    ),
    fields_description="Action/Observation",
    example=(
        'Action: get_weather("New York", 7)'
        "\nObservation: The 7 day forecast for New York is [...]"
    ),
)


def parse_message(m: Message, tools: list[Tool]) -> ToolRequestMessage:  # noqa: C901
    """
    Parse an Act or ReAct Message into a ToolRequestMessage.

    Args:
        m: Input raw message.
        tools: Tools used to confirm a valid tool selection

    Returns:
        Parsed ToolRequestMessage.
    """
    if not m.content:
        raise MalformedMessageError(
            f"{EMPTY_CONTENT_BASE_MSG} of type {type(m).__name__}."
        )

    message_content = m.content
    # strip (and overwrite) up to end of action input
    loc = message_content.find("Action Input:")
    if loc != -1:
        loc = message_content.find("\n", loc)
        message_content = message_content[: loc if loc > 0 else None]
    # we need to override the message too - don't want the model to hallucinate
    m.content = message_content

    action_args: tuple[Any, ...] = ()
    # https://regex101.com/r/qmqZ7Z/1
    action_input = re.search(r"Input:[ \t]*([ \S]*)", m.content)
    # only parse if it takes arguments
    if action_input and action_input.group(1).strip():
        input_str = action_input.group(1).strip()
        # if it has commas and no quotes, it's almost certainly a tuple without
        # parentheses, so we add them
        if "," in input_str and not (
            input_str.startswith("(") and input_str.endswith(")")
        ):
            input_str = f"({input_str})"
        try:
            if input_str.startswith("(") and input_str.endswith(")"):
                # Handle tuples and quoted strings inside
                if '"' not in input_str and "'" not in input_str:
                    # Add quotes around each element within parentheses if they are not already quoted
                    # and if they are not purely numbers. There may exist a killer regex for this
                    # but I am a simple man

                    # just catches things like "1.1".isnumeric() == False
                    # so we can't just use isnumeric
                    def is_number(s: str) -> bool:
                        try:
                            float(s)
                        except ValueError:
                            return False
                        return True

                    input_str = ", ".join(
                        f'"{e.strip()}"' if not is_number(e) else str(e)
                        for e in input_str.strip("()").split(",")
                        if e.strip()
                    )
                    input_str = f"({input_str})"
                eval_result = ast.literal_eval(input_str)
                action_args = (
                    (eval_result,)
                    if not isinstance(eval_result, tuple)
                    else eval_result
                )
            else:
                # Convert to int or float if possible
                try:
                    action_args = (ast.literal_eval(input_str),)
                except (ValueError, SyntaxError):
                    action_args = (input_str,)
        except Exception as exc:
            raise MalformedMessageError(
                f"Action Input {input_str} could not be parsed."
            ) from exc

        if len(action_args) == 1 and isinstance(action_args[0], tuple):
            action_args = action_args[0]

    action = re.search(r"Action:[ \t]*(\S*)", m.content)
    if not action:
        raise MalformedMessageError("Action not emitted.")
    tool_name = action.group(1).strip()
    # have to match up name to tool to line up args in order
    try:
        tool = next(t for t in tools if t.info.name == tool_name)
    except StopIteration as exc:
        raise MalformedMessageError(f"Tool {tool_name} not found in tools.") from exc
    required_parameters = tool.info.parameters.required if tool.info.parameters else []
    if len(action_args) < len(required_parameters):
        raise MalformedMessageError(
            f"Action Input {action_args!r} shorter than {tool.info.name!r} tool's"
            " parameters."
        )

    # Anecdotally we've observed thought also often captures the action
    # NOTE: for Act agents there is no Thought, so the regex will return None
    thought = re.search(r"Thought:[ \t]*(.*)", m.content)
    return ToolRequestMessage(
        content=thought.group(1) if thought else None,
        tool_calls=[ToolCall.from_tool(tool, *action_args)],
    )


class ToolDescriptionMethods(StrEnum):
    """Possible methods of describing the tools."""

    STR = "describe_str"
    XML = "describe_xml"
    JSON = "describe_json"

    def get_prompt_prefix(self) -> str:
        """Get the prefix to put in front of the prompt."""
        if self == self.STR:
            return ""
        if self == self.JSON:
            return "Tools are specified with a JSON schema."
        return "Tools are specified with an XML schema."


class ReActModuleSinglePrompt:
    """An Act or ReAct module built to work with chat models.

    Paper: https://arxiv.org/abs/2210.03629

    The ReAct style is like so, and note Act style has no 'Thought: ' entries:
    System:
        Answer the following questions as best you can. You have access to the following tools:

            {tools}

            Use the following format:

            Thought: you should always think about what to do
            Action: the action to take, should be one of [{tool_names}]
            Action Input: the input to the action
            Observation: the result of the action
            ... (this Thought/Action/Action Input/Observation can repeat N times)
    User:
        {questions}
    Assistant:
        Thought:
        Action:
        Action Input:
    User:
        Observation:
    Assistant:
        Thought:
        Action:
        Action Input:
    ...
    """

    @staticmethod
    def parse_message(m: Message, tools: list[Tool]) -> ToolRequestMessage:
        return parse_message(m, tools)

    async def _create_system_prompt(self, tools: list[Tool]) -> OpResult[str]:
        tool_info = "\n".join([
            getattr(t.info, self._tool_description_method)() for t in tools
        ])
        if prefix := self._tool_description_method.get_prompt_prefix():
            tool_info = f"{prefix}\n{tool_info}"
        tool_names = ", ".join([t.info.name for t in tools])
        return await self.prompt_op(
            schema_type=self._tool_description_method.value,
            tools=tool_info.strip(),
            tool_names=tool_names,
        )

    def __init__(
        self,
        llm_model: dict[str, Any],
        sys_prompt: str = REACT_DEFAULT_SINGLE_PROMPT_TEMPLATE,
        tool_description_method: ToolDescriptionMethods = ToolDescriptionMethods.STR,
    ):
        self.prompt_op = PromptOp(sys_prompt)
        self._tool_description_method = tool_description_method
        llm_model["stop"] = ["Observation:"]
        self.package_msg_op = FxnOp(prepend_sys)
        self.tool_select_module = ParsedLLMCallModule[ToolRequestMessage](
            llm_model=llm_model, parser=self.parse_message
        )

    @property
    def llm_call_op(self) -> LLMCallOp:
        return self.tool_select_module.llm_call_op

    @compute_graph()
    async def __call__(
        self, messages: Iterable[Message], tools: list[Tool]
    ) -> tuple[OpResult[ToolRequestMessage], list[Message]]:
        packaged_msgs = await self.package_msg_op(
            messages, sys_content=await self._create_system_prompt(tools)
        )
        final_result, react_message = await self.tool_select_module(
            packaged_msgs,  # type: ignore[arg-type]
            tools=tools,
        )
        return final_result, [react_message]


def postprocess_and_concat_resoning_msg(
    msgs: Iterable[Message], react_message: Message
) -> Messages:
    return [
        *msgs,
        react_message,
        # We interleave a user message as required by Anthropic's API
        Message(content="Continue..."),
    ]


class ReActModule(ReActModuleSinglePrompt):
    def __init__(
        self,
        llm_model: dict[str, Any],
        sys_prompt: str = REACT_DEFAULT_PROMPT_TEMPLATE,
        tool_description_method: ToolDescriptionMethods = ToolDescriptionMethods.STR,
    ):
        self._tool_description_method = tool_description_method
        llm_model["stop"] = ["Observation:", "Action:"]
        self.llm_config = llm_model
        self._llm_call_op = LLMCallOp()
        self.prompt_op = PromptOp(sys_prompt)
        self.package_msg_op = FxnOp(prepend_sys)
        self.postprocess_reasoning_msg_op = FxnOp(postprocess_and_concat_resoning_msg)

    async def _create_system_prompt(self, tools: list[Tool]) -> OpResult[str]:
        raise NotImplementedError(
            "ReActModule does not implement _create_system_prompt, "
            "since tool descriptions are passed to the API directly "
            "instead of via prompt. Use self.prompt_op instead."
        )

    @property
    def llm_call_op(self) -> LLMCallOp:
        return self._llm_call_op

    @compute_graph()
    async def __call__(
        self, messages: Iterable[Message], tools: list[Tool]
    ) -> tuple[OpResult[ToolRequestMessage], Messages]:
        sys_prompt = await self.prompt_op()

        packaged_msgs = await self.package_msg_op(messages, sys_content=sys_prompt)
        # Ask the LLM to do the reasoning
        reasoning_msg = await self.llm_call_op(
            self.llm_config,
            msgs=packaged_msgs,
            tools=tools,
            tool_choice="none",  # Reasoning shouldn't pick a tool
        )
        # Add the reasoning to messages. Generate the tool selection prompt
        packaged_msgs_with_reasoning = await self.package_msg_op(
            await self.postprocess_reasoning_msg_op(messages, reasoning_msg),
            sys_content=sys_prompt,
        )
        # Ask the LLM to select the tool
        tool_selection_msg = await self.llm_call_op(
            self.llm_config, msgs=packaged_msgs_with_reasoning, tools=tools
        )
        return cast("OpResult[ToolRequestMessage]", tool_selection_msg), [
            # We return the 3 new messages: reasoning (assistant) message,
            # the "continue..." (user) message from user,
            # and tool selection (assistant) message
            *packaged_msgs_with_reasoning.value[-2:],
            tool_selection_msg.value,
        ]



import logging
import logging.config
from typing import Any

from aviary.core import Message, ToolRequestMessage

logger = logging.getLogger(__name__)


def configure_stdout_logs(
    name: str = "root",
    level: int | str = logging.INFO,
    fmt: str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
) -> None:
    """Configure root logger to log to stdout.

    Args:
        name: Optional logger name, if unspecified the 'root' logger is configured.
        level: Log level to be emitted to stdout.
        fmt: Optional format string.
    """
    config: dict[str, Any] = {name: {"level": level, "handlers": ["stdout"]}}
    if name != "root":  # Non-root loggers need to be in a "loggers" key
        config["loggers"] = config
    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {"standard": {"format": fmt}},
            "handlers": {
                "stdout": {
                    "level": "INFO",
                    "formatter": "standard",
                    "class": "logging.StreamHandler",
                    "stream": "ext://sys.stdout",
                },
            },
        }
        | config
    )


def discounted_returns(
    rewards: list[float], terminated: list[bool], discount: float = 1.0
) -> list[float]:
    r"""
    Calculate the discounted returns for a list of rewards, considering termination flags and a discount factor.

    The discounted return represents the future discounted rewards from each time step onwards, taking into account
    whether an episode has terminated at each step.

    The discounted return \( G_t \) is given by:

    .. math::
        G_t = \sum_{k=1}^{\infty} \gamma^{k-1} R_{t+k}

        where:
        - \( G_t \) is the discounted return starting from time step \( t \).
        - \( \gamma \) is the discount factor.
        - \( R_{t+k} \) is the reward received at time step \( t+k \).

    NOTE: this could live in ldp.alg, but it's here to avoid circular imports.

    Args:
        rewards: A list of rewards at each time step.
        terminated: A list of boolean flags indicating whether the episode terminated at each time step.
        discount: Discount factor to apply to future rewards. Defaults to 1.0 which means no discounting is applied.

    Returns:
        A list of discounted returns (rewards to go), with each element representing the
            total discounted reward from that step onwards.

    Example:
        >>> rewards = [1.0, 2.0, 3.0]
        >>> terminated = [False, False, True]
        >>> discounted_returns(rewards, terminated, discount=0.9)
        [5.23, 4.7, 3.0]
    """
    returns = []
    r = 0.0
    for reward, term in zip(reversed(rewards), reversed(terminated), strict=False):
        # 1 - term is 0 if the episode has terminated
        r = reward + discount * r * (1 - term)
        returns.append(r)
    returns.reverse()
    return returns


def format_error_details(error: Exception) -> str:
    """Format detailed error information from an exception.

    Specially handles HTTP errors that have response attributes with status codes
    and JSON details, but works with any exception type.

    Args:
        error: The exception to format

    Returns:
        A formatted error string with available details
    """
    error_details = f"{error!s}"

    if hasattr(error, "response"):
        error_details += f"\nStatus code: {error.response.status_code}"
        try:
            response_data = error.response.json()
            if "detail" in response_data:
                error_details += "\nServer Traceback:\n"
                for line in response_data["detail"].split("\n"):
                    error_details += f"    {line}\n"
        except Exception:
            error_details += f"\nResponse body: {error.response.text}"

    return error_details


def split_message_transitions(list_of_messages: list[Message]) -> list[list[Message]]:
    """
    Break down messages into transitions: [(system, user), (tool request, tool response, env state), ...].

    A block starts with a ToolRequestMessage and ends when a new ToolRequestMessage is encountered.

    Args:
        list_of_messages: The list of messages to break down.

    Returns:
        A list of transitions blocks, corresponding to trajectory transition.
    """
    if not list_of_messages:
        return []

    filtered_messages: list[list[Message]] = []

    # First block: collect all system messages + user message
    system_block: list[Message] = []
    for msg in list_of_messages:
        if msg.role not in {"system", "user"}:
            break
        system_block.append(msg)
    filtered_messages.append(system_block)

    # Process all remaining messages in
    # (ToolRequestMessage, ...) blocks
    current_block: list[Message] = []
    for j in range(len(system_block), len(list_of_messages)):
        msg = list_of_messages[j]
        if isinstance(msg, ToolRequestMessage) and current_block:
            filtered_messages.append(current_block)
            current_block = [msg]
        else:
            current_block.append(msg)

    if current_block:
        filtered_messages.append(current_block)

    return filtered_messages



from __future__ import annotations

import asyncio
import logging
from collections import UserDict
from enum import StrEnum, auto
from typing import Any, Self, cast

from aviary.core import Message
from lmi import LiteLLMModel as LLMModel
from lmi import LLMResult
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    JsonValue,
    RootModel,
    ValidationError,
    model_validator,
)

from ldp.agent import ReActAgent
from ldp.alg.optimizer.opt import Optimizer
from ldp.data_structures import Trajectory
from ldp.graph import LLMCallOp, OpResult, PromptOp

logger = logging.getLogger(__name__)


class APEScoreFn(StrEnum):
    # Use the reward as the APE score (as proposed in the paper).
    # Goal is to maximize this score.
    REWARD = auto()
    # Use the gradient of the output of the PromptOp as the APE score.
    # Goal is to push this to zero.
    GRADIENT = auto()


class _FormatDict(UserDict):
    """Custom dictionary that stores missing items."""

    def __init__(self) -> None:
        super().__init__()
        self.key_set: set[str] = set()

    def __missing__(self, key: str) -> str:
        self.key_set.add(key)
        return key


def get_formatted_variables(s: str) -> set[str]:
    """Returns the set of variables implied by the format string."""
    format_dict = _FormatDict()
    s.format_map(format_dict)
    return format_dict.key_set


class OutputPrompt(BaseModel):
    prompt: str = Field(description="Prompt for language model")


class Example(BaseModel):
    input: JsonValue
    output: JsonValue
    score: float


ExampleList = RootModel[list[Example]]


class APEOpt(BaseModel, Optimizer):
    """
    Basic optimizer that acts as an Automatic Prompt Engineer (APE).

    Paper: https://openreview.net/pdf?id=92gvk82DE-

    Details:
    - This implements the "forward mode generation" strategy.
    - The score function used is the gradient (float) at the output of the
      PromptOp being optimized. A zero gradient means the prompt was "good",
      and a non-zero gradient means we can learn from the prompt.
    - Possible improvements include:
        - Extending the score function to the LLM result's logprobs
        - Iterating with Monte Carlo Search
        - Use of memory for further example augmentation
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    # Configuration
    max_examples: int | None = Field(
        default=50,  # Comes from APE paper section 5.2
        description=(
            "Max number of examples to include in the below query_prompt, or None for"
            " no limit. The paper mentions that more examples produce better prompt"
            " proposals."
        ),
    )
    system_prompt: str = (
        "We are optimizing prompts for a language model."
        " The model sees a prompt, an input, and then generates an output."
    )
    query_prompt: str = (
        "Here are correct example outputs that the language model"
        " and prompt should produce:\n{good_examples}"
        '\n\nThe current prompt is: "{prompt}"'
        "\n\nWhich resulted in the following incorrect input, output, and {score}:"
        "\n{examples}\n\nRevise the current prompt to improve the outputs."
        " Your proposed prompt should be concise, correct, and specify the desired"
        " output format."
    )
    llm: LLMModel = Field(
        default_factory=LLMModel,
        description=(
            "LLM used to update the prompt inside the PromptOp. The paper mentions that"
            " larger models produce better prompt proposals."
        ),
    )

    prompt_op: PromptOp = Field(description="PromptOp to be optimized.")
    llm_call_op: LLMCallOp = Field(description="LLMCallOp to be optimized.")

    score_fn: APEScoreFn = APEScoreFn.REWARD
    good_reward_threshold: float | None = Field(
        default=None,
        description=(
            "If using reward as the score_fn, then a good example is defined by "
            "reward>=good_reward_threshold."
        ),
    )
    reward_discount: float = 1.0

    # State
    examples: list[Example] = Field(default_factory=list)
    good_examples: list[Example] = Field(default_factory=list)
    steps: int = 0
    trace: list[str] = Field(
        default_factory=list, description="History of prompts used."
    )

    @model_validator(mode="after")
    def validate_score_fn(self):
        if self.score_fn == APEScoreFn.REWARD:
            if self.good_reward_threshold is None:
                raise ValueError(
                    "good_reward_threshold must be set if using reward as the score"
                    " function"
                )
            self._score_str = "rewards"
        elif self.score_fn == APEScoreFn.GRADIENT:
            # The gradient into the prompt op is the (signed) backpropagated error, and "gradient" would
            # be confusing to the model in the prompt.
            self._score_str = "errors"
        else:
            raise ValueError(f"Invalid score function {self.score_fn}")

        return self

    def model_post_init(self, __context: Any) -> None:  # noqa: PYI063
        if self.prompt_op.prompt not in self.trace:
            self.trace.append(self.prompt_op.prompt)

        # Make sure updates are not run concurrently
        self._update_lock = asyncio.Lock()

    @classmethod
    def from_agent(cls, agent: ReActAgent, **kwargs) -> Self:
        return cls(
            llm_call_op=agent._react_module.llm_call_op,
            prompt_op=agent._react_module.prompt_op,
            **kwargs,
        )

    def aggregate_trajectory(self, trajectory: Trajectory) -> None:
        if trajectory.failed:
            return

        if self.score_fn == APEScoreFn.REWARD:
            d_returns = trajectory.compute_discounted_returns(self.reward_discount)

        for i_step, step in enumerate(trajectory.steps):
            action = cast("OpResult", step.action)

            if self.score_fn == APEScoreFn.GRADIENT:
                prompt_op_result, *extra = action.get_upstream_results(self.prompt_op)
                assert not extra, "APE only supports one prompt call per run"

            for op_result in action.get_upstream_results(self.llm_call_op):
                result = cast(
                    "LLMResult | None",
                    self.llm_call_op.ctx.get(op_result.call_id, "result"),
                )
                if result is None or not result.messages or not result.prompt:
                    continue
                # (x: first prompt's user message's content, y: AI response's content)
                x = next(
                    # m is a Message with a result of the LLM. Which completes only strings.
                    # and we checked the result exists above.
                    cast("str", m.content)
                    for m in result.prompt
                    if (isinstance(m, Message) and m.role == "user")
                )
                y = cast("str", result.messages[0].content)

                if self.score_fn == APEScoreFn.GRADIENT:
                    score = self.prompt_op.ctx.get(
                        prompt_op_result.call_id, "grad_output", default=None
                    )
                    if score is None:
                        # backprop did not reach this op call - move on
                        continue
                    is_good = score == 0
                else:
                    score = d_returns[i_step]  # pylint: disable=possibly-used-before-assignment
                    is_good = score >= cast("float", self.good_reward_threshold)

                example = Example(input=x, output=y, score=score)
                (self.good_examples if is_good else self.examples).append(example)

    async def update(self) -> None:
        async with self._update_lock:
            if not self.examples:
                raise ValueError("No examples to update the prompt with.")

            new_p = await self._get_updated_prompt(
                self.examples, self.good_examples, self.prompt_op.prompt
            )
            # Check any template vars remain, and if some were added or
            # lost, discard this new prompt
            if new_p != self.prompt_op.prompt and get_formatted_variables(
                new_p
            ) != get_formatted_variables(self.prompt_op.prompt):
                logger.warning(
                    "Update broke prompt templating."
                    f"\n\nNew prompt:\n{new_p}"
                    f"\n\nPrior prompt:\n{self.prompt_op.prompt}"
                )
            else:
                if new_p == self.prompt_op.prompt:
                    logger.warning("Update did not change the prompt.")
                self.examples.clear()
                self.prompt_op.prompt = new_p
                self.trace.append(new_p)
            self.steps += 1

    def _prepare_examples(self, examples: list[Example]) -> str:
        if not examples:
            return ""
        if self.max_examples and len(examples) > self.max_examples:
            if self.score_fn == APEScoreFn.GRADIENT:
                # Return examples closest to decision boundary,
                # aka ones with the lowest L1-normalized error
                # NOTE: this pairs with our setting of Example.score = PromptOp's output
                # gradient inside the update method, so examples with error values closer to
                # 0 are defined to be higher quality
                examples = sorted(examples, key=lambda e: abs(e.score))[
                    : self.max_examples
                ]
            else:
                # In reward mode, we want to show the examples with the highest reward, per the paper
                # TODO: consider whether uniform sampling is better
                examples = sorted(examples, key=lambda e: -e.score)[: self.max_examples]
        return ExampleList.model_validate(examples).model_dump_json()

    async def _get_updated_prompt(
        self, examples: list[Example], good_examples: list[Example], prompt: str
    ) -> str:
        messages = [
            Message(
                role="system",
                content=self.system_prompt,
            ),
            Message(
                role="user",
                content=self.query_prompt.format(
                    examples=self._prepare_examples(examples),
                    good_examples=self._prepare_examples(good_examples),
                    prompt=prompt,
                    score=self._score_str,
                ),
            ),
            Message(
                role="assistant",
                content=OutputPrompt(prompt=prompt).model_dump_json(indent=2),
            ),
            Message(
                content=(
                    "You responded without changing the prompt. Don't forget to revise"
                    " the prompt."
                )
            ),
        ]
        result = await self.llm.call_single(messages, output_type=OutputPrompt)
        message_content = cast(
            "str", cast("list[Message]", result.messages)[-1].content
        )
        try:
            return OutputPrompt.model_validate_json(message_content).prompt
        except ValidationError:
            return prompt
